import subprocess
import networkx as nx
import re
import os
from .config.path import *
from .config.config import *
from typing import Dict
from .argument import Arguments




class Graph:
    def __init__(self, benchmark_name, is_clean: bool = False, address: str = None):
        """
        takes in a circuit and creates a networkx graph out of it
        :param benchmark_name: the input benchmark in gv format
        :param is_clean: leave empty for now
        """
        self.__graph_name = benchmark_name
        folder, extension = INPUT_PATH['gv'] # input/gv/
        self.__graph_in_path = f'{folder}/{benchmark_name}.{extension}'

        folder, extension = OUTPUT_PATH['gv'] # output/gv
        self.__graph_out_path = f'{folder}/{benchmark_name}.{extension}'

        folder, extension = OUTPUT_PATH['dot'] # output/gv
        self.__dot_in_path = f'{folder}/{benchmark_name}.{extension}'

        folder, extension = OUTPUT_PATH['ver'] # output/ver
        self.__verilog_in_path = f'{folder}/{benchmark_name}.{extension}'

        # print(f'Converting the graph into a NeworkX object')


        self.__graph = self.import_graph(address)

        self.__sorted_node_list = None

        self.__is_clean = is_clean


        if not self.is_clean:
            # print(f'We are cleaning!')
            self.clean_graph()

        # print(f'After cleaning!')
        # for n in self.graph.nodes:
        #     print(f'{n = }, {self.graph.nodes[n] = }')
        self.sort_graph()
        self.remove_output_outgoing_edges(self)
        self.__input_dict = self.sort_dict(self.extract_inputs())
        self.__output_dict = self.sort_dict(self.extract_outputs())
        self.__gate_dict = self.sort_dict(self.extract_gates())
        self.__constant_dict = self.sort_dict(self.extract_constants())




        self.__num_inputs = len(self.__input_dict)
        self.__num_outputs = len(self.__output_dict)
        self.__num_gates = len(self.__gate_dict)
        self.__num_constants = len(self.__constant_dict)

    @property
    def name(self):
        return self.__graph_name

    @property
    def is_clean(self):
        return self.__is_clean

    def set_is_clean(self, is_clean: bool):
        self.__is_clean = is_clean

    @property
    def in_path(self):
        return self.__graph_in_path

    @property
    def out_path(self):
        return self.__graph_out_path

    @property
    def verilog_in_path(self):
        return self.__verilog_in_path

    @property
    def dot_in_path(self):
        return self.__dot_in_path

    @property
    def graph(self):
        return self.__graph

    def set_graph(self, graph):
        self.__graph = graph

    @property
    def input_dict(self):
        return self.__input_dict

    def set_input_dict(self, input_dict):
        self.__input_dict = input_dict

    @property
    def output_dict(self):
        return self.__output_dict

    def set_output_dict(self, output_dict):
        self.__output_dict = output_dict

    @property
    def gate_dict(self):
        return self.__gate_dict

    def set_gate_dict(self, gate_dict):
        self.__gate_dict = gate_dict

    @property
    def constant_dict(self):
        return self.__constant_dict

    def set_constant_dict(self, constant_dict):
        self.__constant_dict = constant_dict

    @property
    def num_inputs(self):
        return self.__num_inputs

    def set_num_inputs(self, input_count):
        self.__num_inputs = input_count

    @property
    def num_outputs(self):
        return self.__num_outputs

    def set_num_outputs(self, output_count):
        self.__num_outputs = output_count

    @property
    def num_gates(self):
        return self.__num_gates

    def set_num_gates(self, gate_count):
        self.__num_gates = gate_count

    @property
    def sorted_node_list(self):
        return self.__sorted_node_list


    def remove_output_outgoing_edges(self):
            for n in self.graph.nodes:
                if self.is_cleaned_po(n):
                    if len(list(self.graph.successors(n))) > 0:

                        if len(list(self.graph.predecessors(n))) > 1:
                            raise Exception(Fore.RED + f'Output {n} has more than 1 predecessors!' + Style.RESET_ALL)

                        pred = list(self.graph.predecessors(n))[0]
                        for s in list(self.graph.successors(n)):
                            self.graph.remove_edge(n, s)
                            self.graph.add_edge(pred, s)

    def set_sorted_node_list(self, sorted_node_list):
        self.__sorted_node_list = sorted_node_list

    def sort_dict(self, this_dict: Dict) -> Dict:
        sorted_dict: type(this_dict) = {}
        this_dict_ids = list(this_dict.keys())
        # print(f'{this_dict_ids = }')

        this_dict_ids.sort()
        # print(f'{this_dict_ids = }')


        for i in this_dict_ids:
            sorted_dict[i] = this_dict[i]

        return sorted_dict

    def extract_inputs(self):
        # print(f'Extracting inputs...')
        input_dict = {}
        idx = 0
        for n in self.graph.nodes():
            if self.is_pi(n):
                idx = re.search('\d+', n).group()
                input_dict[int(idx)] = n
                # idx += 1
        return input_dict

    def extract_outputs(self):
        # print(f'Extracting outputs...')
        output_dict = {}
        idx = 0
        for n in self.graph.nodes():
            if self.is_po(n):
                idx = re.search('\d+', n).group()
                output_dict[int(idx)] = n
                # idx += 1
        return output_dict

    def extract_gates(self):
        gate_dict = {}
        idx = 0
        for n in self.graph.nodes:
            if self.is_cleaned_gate(n):
                idx = re.search('\d+', n).group()
                gate_dict[int(idx)] = n
                # idx += 1
        return gate_dict

    def extract_constants(self):
        constant_dict = {}
        idx = 0
        for n in self.graph.nodes:
            if self.is_constant(n):
                idx = re.search('\d+', n).group()
                constant_dict[int(idx)] = n
                # idx += 1
        return constant_dict

    # methods
    def create_sort_map(self) -> 'dict[str, str]':
        """
        sorts the nodes in topological order and returns them in a form of a list
        :return: a sorted list of node labels
        """
        sorted_node_list = list(nx.topological_sort(self.graph))
        sort_map = {}
        idx = 0
        for n in sorted_node_list:
            if re.search(r'in\d+', n):
                sort_map[n] = n
            elif re.search(r'out\d+', n):
                sort_map[n] = n
            elif re.search(r'g\d+', n):
                sort_map[n] = f'g{idx}'
                idx += 1

        return sort_map

    def sort_graph(self):
        sort_map = self.create_sort_map()
        self.set_graph(nx.relabel_nodes(self.graph, sort_map))
        self.set_sorted_node_list(list(nx.topological_sort(self.graph)))

        tmp_graph = nx.DiGraph()
        tmp_graph.add_nodes_from(self.sorted_node_list)
        tmp_graph.add_edges_from(self.graph.edges())
        for (k, v) in self.graph.nodes(data=True):
            for attr in v.keys():
                tmp_graph.nodes[k][attr] = v[attr]

        self.set_graph(tmp_graph)
        # print(f'{tmp_graph.nodes = }')

    # TODO
    # Deprecated
    def count_iog(self):
        """
        counts the number of primary inputs and outputs, and gates and sets the corresponding fields
        :return: Nothing
        """
        count_i = 0
        count_o = 0
        count_g = 0
        for n in self.graph.nodes:
            if re.search(r'in\d+', n) and re.search(r'in\d+', self.graph.nodes[n]['label']):
                count_i += 1
            elif re.search(r'out\d+', n) and re.search(r'out\d+', self.graph.nodes[n]['label']):
                count_o += 1
            elif re.search(r'g\d+', n) and re.search(POSSIBLE_GATES, self.graph.nodes[n]['label']):
                count_g += 1
        self.set_num_inputs(count_i)
        self.set_num_outputs(count_o)
        self.set_num_gates(count_g)

    def convert_verilog_to_gv(self):
        yosys_command = f"""
            read_verilog {self.verilog_in_path}
            show -prefix {self.out_path[:-3]} -format gv
            """
        with open(f'yosys_graph.log', 'w') as y:
            subprocess.call([YOSYS, '-p', yosys_command], stdout=y)
        self.fix_direction()

    def fix_direction(self):
        dot_command = f'{DOT} {self.dot_in_path} -Grankdir=TB -o {self.out_path}'
        subprocess.call([dot_command], shell=True)

    def clean_graph(self):
        # for n in self.graph.nodes:
        #     if LABEL not in self.graph.nodes[n]:
        #         print(f'{n}, {self.graph.nodes[n] = }')

        # for n in self.graph.nodes:
        #     if LABEL not in self.graph.nodes[n]:
        #         print(f'{n}, {self.graph.nodes[n] = }')
        # exit()
        self.merge_points_into_source()
        self.clean_input_labels()
        self.clean_output_labels()
        self.clean_wire_labels()
        self.clean_gate_labels()
        self.clean_constant_labels()
        self.merge_buffers_into_gates()
        self.merge_wires_into_gates()
        self.relabel_nodes()
        self.delete_extra_fields()
        self.set_is_clean(True)
        self.count_iog()
        self.set_input_dict(self.extract_inputs())
        self.set_output_dict(self.extract_outputs())
        self.set_gate_dict(self.extract_gates())
        self.set_constant_dict(self.extract_constants())


    def delete_extra_fields(self):
        for n in self.graph.nodes:
            for field in REDUNDANT_FIELDS:
                if field in self.graph.nodes[n].keys():
                    self.graph.nodes[n].pop(field)
        # delete all edge attributes
        for (n1, n2, d) in self.graph.edges(data=True):
            d.clear()

    def clean_input_labels(self):
        self.delete_extra_fields()
        for n in self.graph.nodes:
            if self.is_pi(n):
                idx = re.search('\d+', self.graph.nodes[n][LABEL]).group()  # pi7 => idx=7
                self.graph.nodes[n][LABEL] = f'in{idx}'
                self.graph.nodes[n][SHAPE] = f'circle'

    def clean_output_labels(self):
        for n in self.graph.nodes:
            if self.is_po(n):

                idx = re.search('\d+', self.graph.nodes[n]['label']).group()  # po12 => idx=12
                self.graph.nodes[n]['label'] = f'out{idx}'
                self.graph.nodes[n]['shape'] = f'doublecircle'

    def clean_wire_labels(self):
        for n in self.graph.nodes:
            if self.is_wire(n):
                idx = re.search('\d+', self.graph.nodes[n]['label']).group()  # n317 => idx=317
                self.graph.nodes[n]['label'] = f'g{idx}'
                self.graph.nodes[n]['shape'] = f'square'

    def clean_gate_labels(self):
        for n in self.graph.nodes:
            if self.is_gate(n):
                if re.search(POSSIBLE_GATES, self.graph.nodes[n]['label']):
                    cur_gate = re.search(POSSIBLE_GATES, self.graph.nodes[n]['label']).group()
                    self.graph.nodes[n]['label'] = cur_gate
                    self.graph.nodes[n]['shape'] = 'invhouse'


    def clean_constant_labels(self):
        # g16[label = "F", shape = circle, fillcolor = white]
        for n in self.graph.nodes:
            if self.is_constant(n):
                const_value = re.search(f'{CONST_0}|{CONST_1}', self.graph.nodes[n][LABEL]).group()
                self.graph.nodes[n][LABEL] = f"{'FALSE' if const_value == CONST_0 else 'TRUE'}"
                self.graph.nodes[n][SHAPE] = f'square'

    def merge_buffers_into_gates(self):
        tmp_graph = self.graph.copy(as_view=False)
        for e in self.graph.edges:
            if self.graph.nodes[e[1]]['label'] == 'BUF':
                src_node = e[0]
                des_node = e[1]
                tmp_graph = nx.contracted_nodes(tmp_graph, src_node, des_node, self_loops=False)
        self.set_graph(tmp_graph)

    def merge_points_into_source(self):
        tmp_graph = self.graph.copy(as_view=False)
        for e in self.graph.edges:
            if self.graph.nodes[e[1]][SHAPE] == 'point':
                # print(f'{e[1] = }')
                src_node = e[0]
                des_node = e[1]
                tmp_graph = nx.contracted_nodes(tmp_graph, src_node, des_node, self_loops=False)
        self.set_graph(tmp_graph)

    def merge_wires_into_gates(self):
        tmp_graph = self.graph.copy(as_view=False)
        for e in self.graph.edges:
            # if (re.search(POSSIBLE_GATES, self.graph.nodes[e[0]]['label']) and
            #         re.search('g\d+', self.graph.nodes[e[1]]['label'])):
            if re.search('g\d+', self.graph.nodes[e[1]]['label']):
                src_node = e[0]
                des_node = e[1]
                tmp_graph = nx.contracted_nodes(tmp_graph, src_node, des_node, self_loops=False)
        self.set_graph(tmp_graph)


    def relabel_nodes(self):
        tmp_graph = self.graph.copy()
        gate_idx = 0
        for n in self.graph.nodes:
            if self.is_cleaned_pi(n):
                old_name = n
                new_name = self.graph.nodes[n]['label']

                mapping = {old_name: new_name}
                tmp_graph = nx.relabel_nodes(tmp_graph, mapping)
            elif self.is_cleaned_po(n):
                old_name = n
                new_name = self.graph.nodes[n]['label']

                mapping = {old_name: new_name}
                tmp_graph = nx.relabel_nodes(tmp_graph, mapping)
            elif self.is_merged_gate(n):
                old_name = n
                new_name = f'g{gate_idx}'
                gate_idx += 1

                mapping = {old_name: new_name}
                tmp_graph = nx.relabel_nodes(tmp_graph, mapping)
            elif self.is_cleaned_gate(n):
                old_name = n
                new_name = f'g{gate_idx}'
                gate_idx += 1

                mapping = {old_name: new_name}
                tmp_graph = nx.relabel_nodes(tmp_graph, mapping)
            elif self.is_constant(n):
                old_name = n
                new_name = f'g{gate_idx}'
                gate_idx += 1

                mapping = {old_name: new_name}
                tmp_graph = nx.relabel_nodes(tmp_graph, mapping)
            else:
                print('WARNING! No mapping needed!')

        self.set_graph(tmp_graph)


    def is_cleaned_pi(self, node):
        if not self.is_constant(node):
            if re.search(r'in\d+', self.graph.nodes[node][LABEL]):
                return True
            else:
                return False
        else:
            return False

    def is_pruned_pi(self, node):
        if not self.is_constant(node):
            if (re.search(r'in\d+', self.graph.nodes[node][LABEL]) and
                    PRUNED in self.graph.nodes[node]):
                return True
            else:
                return False
        else:
            return False

    def is_pi(self, node):
        # print(f'{node = }')
        # print(f'{self.graph.nodes[node] = }')

        if not self.is_constant(node):
            if re.search(r'pi\d+', self.graph.nodes[node][LABEL]) or \
                self.graph.nodes[node][LABEL].startswith('in'):
                return True
            else:
                return False
        else:
            return False

    def is_cleaned_po(self, node):
        if not self.is_constant(node):
            if re.search(r'out\d+', self.graph.nodes[node][LABEL]):

                return True
            else:
                return False
        else:
            return False

    def is_po(self, node):
        if not self.is_constant(node):
            if re.search(r'po\d+', self.graph.nodes[node][LABEL]) or \
               self.graph.nodes[node][LABEL].startswith('out'):
                return True
            else:
                return False
        else:
            return False

    def is_wire(self, node):
        if not self.is_constant(node):
            if (not self.is_pi(node) and not self.is_po(node) and not self.is_gate(node) and
                    re.search('diamond', self.graph.nodes[node]['shape'])):
                return True
            else:
                return False
        else:
            return False

    def is_cleaned_gate(self, node):
        if not self.is_constant(node):
            if (re.search('invhouse', self.graph.nodes[node][SHAPE]) and
                    'contraction' not in self.graph.nodes[node].keys() and 'contractions' not in self.graph.nodes[
                        node]):
                return True
            elif (re.search('invhouse', self.graph.nodes[node][SHAPE]) and re.search(r'g\d+', node) and
                  'contraction' not in self.graph.nodes[node].keys() and 'contractions' not in self.graph.nodes[
                      node]):
                return True
            else:
                return False
        else:
            return False

    def is_pruned_gate(self, node):
        if not self.is_constant(node):
            if self.is_cleaned_gate(node) and PRUNED in self.graph.nodes[node]:
                return True
            else:
                return False
        else:
            return False

    def is_gate(self, node):
        if not self.is_constant(node):
            if re.search('record', self.graph.nodes[node]['shape']):
                return True
            else:
                return False
        else:
            return False

    def is_constant(self, node):
        if (LABEL in LABEL in self.graph.nodes[node]):

            if (SHAPE not in self.graph.nodes[node] and
                re.search(f'{CONST_0}|{CONST_1}', self.graph.nodes[node][LABEL])):
                return True
            elif re.search(r'(TRUE|FALSE)', self.graph.nodes[node][LABEL]):
                regex_constant = re.search(f'', self.graph.nodes[node][LABEL]).group()
                return True
        else:
            return False

    def is_cleaned_constant(self, node):
        if re.search(r'(TRUE|FALSE)', self.graph.nodes[node][LABEL]):
            return True
        else:
            return False

    def is_buff(self, node):
        if not self.is_constant(node):
            if (re.search('BUF', self.graph.nodes[node]['label']) and
                    re.search('box', self.graph.nodes[node]['shape'])):
                return True
            else:
                return False
        else:
            return False

    def is_merged_gate(self, node):
        if not self.is_constant(node):
            if (re.search(POSSIBLE_GATES, self.graph.nodes[node]['label']) and
                    re.search('invhouse', self.graph.nodes[node]['shape']) and
                    ('contraction' in self.graph.nodes[node].keys() or 'contractions' in self.graph.nodes[
                        node].keys())):
                return True
            else:
                return False
        else:
            return False

    def import_graph(self, address: str = None):
        """
        :param address: the location of the gv file. default is self.out_path (output/gv/)
        :return: the graph object of networkx
        """

        if address:
            DG = nx.drawing.nx_agraph.read_dot(address)
        else:
            DG = nx.drawing.nx_agraph.read_dot(self.out_path)
        return DG

    def export_graph(self):
        with open(self.out_path, 'w') as f:
            f.write(f"strict digraph \"\" {{\n")
            for n in self.graph.nodes:
                self.export_node(n, f)
            for e in self.graph.edges:
                self.export_edge(e, f)
            f.write(f"}}\n")

    def export_node(self, n, file_handler: 'class _io.TextIOWrapper'):
        # in3	[label=in3,shape=circle];
        if self.is_cleaned_pi(n) or self.is_cleaned_po(n):

            line = f"{n} [label=\"{self.graph.nodes[n]['label']}\", shape={self.graph.nodes[n]['shape']}];\n"
        elif self.is_cleaned_gate(n):

            line = f"{n} [label=\"{self.graph.nodes[n]['label']}\\n{n}\", shape={self.graph.nodes[n]['shape']}];\n"
        elif self.is_cleaned_constant(n):

            line = f"{n} [label=\"{self.graph.nodes[n]['label']}\\n{n}\", shape={self.graph.nodes[n]['shape']}];\n"
        else:
            print('WARNING!!! found a node that is not a PI, PO, WIRE, CONSTANT, GATE')
            raise Exception('NodeTypeError')

        file_handler.write(line)

    def export_edge(self, e: tuple, file_handler):
        # in3 -> g2;
        src, des = e
        line = f"{src} -> {des};\n"
        file_handler.write(line)

    def convert_to_verilog(self):
        pass

    def __repr__(self):
        return f'An object of class Graph\n' \
               f'{self.name = }\n' \
               f'{self.in_path = }\n' \
               f'{self.out_path = }\n' \
               f'{self.num_inputs = }\n' \
               f'{self.num_outputs = }\n' \
               f'{self.num_gates = }\n'



import re
import sys
import random
import os
import argparse
import subprocess
import time
import yaml
from .config.config import *


def create_testbench(input_file, output_file, num, yosys):
    f = open(output_file, 'w')
    modulename, port_list, inp, n_inputs, out, n_outputs = module_info(input_file, yosys)

    f.write("module " + modulename + "_tb;\n")
    f.write('reg [' + str(n_inputs - 1) + ':0] pi;\n')
    f.write('wire [' + str(n_outputs - 1) + ':0] po;\n')
    f.write(modulename + ' dut(')

    first = True
    inp_count = 0
    out_count = 0

    for i in port_list:
        if not first:
            f.write(',')
        first = False

        if i in inp:
            if inp[i] > 1:
                f.write(' pi[{}:{}] '.format(inp_count + inp[i] - 1, inp_count))
            else:
                f.write(' pi[{}] '.format(inp_count))
            inp_count += inp[i]

        elif i in out:
            if out[i] > 1:
                f.write(' po[{}:{}] '.format(out_count + out[i] - 1, out_count))
            else:
                f.write(' po[{}] '.format(out_count))
            out_count += out[i]

        else:
            print('[Error] Port {} is not defined as input or output.'.format(i))
            sys.exit(0)

    f.write(');\n')

    f.write("initial\n")
    f.write("begin\n")
    if n_inputs >= 17:
        j = 1
        while j <= int(num):
            f.write('# 1  pi=' + str(n_inputs) + '\'b')
            i = 1
            while i <= n_inputs:
                n = random.randint(0, 1)
                i += 1
                f.write(str(n))
            f.write(';\n')
            f.write("#1 $display(\"%b\", po);\n")
            j += 1
    else:
        for j in range(2 ** n_inputs):
            f.write('# 1  pi=' + str(n_inputs) + '\'b')
            f.write('{0:0>{1}}'.format(str(bin(j))[2:], n_inputs))
            f.write(';\n')
            f.write("#1 $display(\"%b\", po);\n")

    f.write("end\n")
    f.write("endmodule\n")

    f.close()


def module_info(fname: str):
    tmp = time.strftime('%Y_%m_%d-%H_%m_%s') + '.v'
    yosys_command = 'read_verilog ' + fname + '; synth -flatten; opt; opt_clean; techmap; write_verilog ' + tmp + ';\n'
    subprocess.call([YOSYS, '-p', yosys_command], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)


    tmp_file = open(tmp)
    inp = {}
    inp_count = 0
    out = {}
    out_count = 0
    modulename = None
    line = tmp_file.readline()
    while line:
        tokens = re.split('[ ()]', line.strip().strip(';').strip())

        if len(tokens) > 0 and tokens[0] == 'module' and modulename is None:
            modulename = tokens[1]
            port_list = re.split('[,()]', line.strip().strip(';').strip())[1:]
            port_list = [s.strip() for s in port_list if s.strip() != '']

        if len(tokens) == 2 and (tokens[0] == 'input' or tokens[0] == 'output'):
            if tokens[0] == 'input':
                inp[tokens[1]] = 1
                inp_count += 1
            if tokens[0] == 'output':
                out[tokens[1]] = 1
                out_count += 1

        if len(tokens) == 3 and (tokens[0] == 'input' or tokens[0] == 'output'):
            range_str = tokens[1][1:-1].split(':')
            range_int = list(map(int, range_str))
            length = max(range_int) - min(range_int) + 1
            if tokens[0] == 'input':
                inp[tokens[2]] = length
                inp_count += length
            if tokens[0] == 'output':
                out[tokens[2]] = length
                out_count += length

        line = tmp_file.readline()

    tmp_file.close()

    os.remove(tmp)

    return modulename, port_list, inp, inp_count, out, out_count


def main():
    app_path = os.path.dirname(os.path.realpath(__file__))

    # Parse command-line args
    parser = argparse.ArgumentParser(
        description='BLASYS -- Approximate Logic Synthesis Using Boolean Matrix Factorization')
    parser.add_argument('-i', help='Input verilog file', required=True, dest='input')
    parser.add_argument('-o', help='Output testbench file', required=True, dest='output')
    parser.add_argument('-n', help='Number of test vectors', type=int, default=10000, dest='number')
    args = parser.parse_args()

    print('Start creating testbench ...')

    # Load path to yosys, lsoracle, iverilog, vvp, abc
    with open(os.path.join(app_path, 'config', 'params.yml'), 'r') as config_file:
        config = yaml.safe_load(config_file)

    create_testbench(args.input, args.output, args.number, config['yosys'])


if __name__ == '__main__':
    main()



import os
import re
import natsort

from typing import List
from copy import deepcopy

from stats import Stats
from specs import Specs
from .config.path import *
from .utils import *
import matplotlib
import matplotlib.pyplot as plt
import csv
import numpy as np

class Result:
    def __init__(self, specifications: Specs):
        self.__benchmark: str = get_pure_name(specifications.benchmark)
        self.__approximate_benchmark: str = get_pure_name(specifications.approximate_benchmark)
        self.__experiment: str = specifications.experiment
        self.__strategy: str = specifications.strategy
        self.__optimization: str = specifications.optimization
        self.__metric: str = specifications.metric
        self.__precision: int = specifications.precision

        # Todo: Make this part a bit more generic!
        # Sometimes we want the result of only one benchmark
        # Sometimes we want the result of a family of benchmarks or a whole experiment
        self.__in_paths: List[str] = self.find_input_report_paths()
        self.__reports: List[Stats] = self.import_reports()

    @property
    def benchmark(self):
        return self.__benchmark

    @property
    def approximate_benchmark(self):
        return self.__approximate_benchmark

    @property
    def experiment(self):
        return self.__experiment

    @property
    def strategy(self):
        return self.__strategy

    @property
    def optimization(self):
        return self.__optimization

    @property
    def metric(self):
        return self.__metric

    @property
    def precision(self):
        return self.__precision

    @property
    def reports(self):
        return self.__reports

    @reports.setter
    def reports(self, reports: List[Stats]):
        self.__reports = deepcopy(reports)

    @property
    def in_paths(self):
        return self.__in_paths

    @in_paths.setter
    def in_paths(self, list):
        self.__in_paths = list

    # TODO
    # fix naming: include num_samples

    def find_relevant_csv(self):
        relevant_csv = []
        folder, extension = INPUT_PATH['report']

        if self.metric == WRE:
            if (self.optimization == OPTIMIZE or self.optimization == MAXIMIZE) and (self.strategy != BISECTION):
                cur_dir = f'{folder}/{self.benchmark}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{self.optimization}'
            else:
                cur_dir = f'{folder}/{self.benchmark}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}'
        else:
            if self.optimization == OPTIMIZE or self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                cur_dir = f'{folder}/{self.benchmark}_{self.experiment}_{self.metric}_{self.strategy}_{self.optimization}'
            else:
                cur_dir = f'{folder}/{self.benchmark}_{self.experiment}_{self.metric}_{self.strategy}'

        all_csv = [f for f in os.listdir(cur_dir)]
        for f in all_csv:
            if f.endswith(extension):
                relevant_csv.append(f'{cur_dir}/{f}')

        relevant_csv = natsort.natsorted(relevant_csv)  # sorting the files alphabetically in descending order

        return relevant_csv

    def find_input_report_paths(self):
        input_paths = []
        folder, extension = INPUT_PATH['report']
        if self.experiment == SINGLE:
            input_paths = self.find_relevant_csv()
        elif self.experiment == QOR:
            if self.metric == WRE:
                if self.optimization == OPTIMIZE or self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                    input_paths.append(
                        f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{self.optimization}.{extension}')
                else:
                    input_paths.append(
                        f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}.{extension}')
            else:
                if self.optimization == OPTIMIZE or self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                    input_paths.append(
                        f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_{self.strategy}_{self.optimization}.{extension}')
                else:
                    input_paths.append(
                        f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_{self.strategy}.{extension}')

        elif self.experiment == RANDOM:
            pass

        return input_paths

    def import_reports(self):
        reports: List = []
        for report_path in self.in_paths:
            reports.append(Stats(report_path))
        return reports

    def find_relevant_paths(self, folder: str, extension: str):

        all = [f for f in os.listdir(folder)]
        labelling_folders = []
        qor_paths = []

        for f in all:
            if re.search(self.metric, f):
                if f.endswith(extension):
                    qor_paths.append(f'{folder}/{f}')
                else:
                    labelling_folders.append(f'{folder}/{f}')

        labelling_paths = []
        for fold in labelling_folders:
            cur_files = [f for f in os.listdir(fold)]
            for c_file in cur_files:
                if c_file.endswith(extension):
                    labelling_paths.append(f'{fold}/{c_file}')

        relevant_paths = labelling_paths + qor_paths
        self.in_paths = []
        self.in_paths = relevant_paths

    def extract_fields_for_three_strategies(self):

        monotonic_files = []
        kind_files = []
        bisection_files = []

        for report_path in self.in_paths:
            if re.search(MONOTONIC, report_path):
                monotonic_files.append(report_path)
            elif re.search(KIND_BISECTION, report_path):
                kind_files.append(report_path)
            elif re.search(BISECTION, report_path):
                bisection_files.append(report_path)



        monotonic_files = natsort.natsorted(monotonic_files)
        kind_files = natsort.natsorted(kind_files)
        bisection_files = natsort.natsorted(bisection_files)

        count = 0
        for i in range(len(monotonic_files)):
            mono = monotonic_files[i].replace(MONOTONIC, '')
            kind = kind_files[i].replace(KIND_BISECTION, '')
            bisec = bisection_files[i].replace(BISECTION, '')
            if mono == kind == bisec:
                count += 1

        if count == len(monotonic_files):
            print(f'TEST -> PASS')


        monotonic_reports = []
        kind_reports = []
        bisection_reports = []

        for report_path in monotonic_files:
            monotonic_reports.append(Stats(report_path))
        for report_path in kind_files:
            kind_reports.append(Stats(report_path))
        for report_path in bisection_files:
            bisection_reports.append(Stats(report_path))

        return monotonic_reports, kind_reports, bisection_reports


    def collect_reports_for_three_strategies(self):
        folder, extension = OUTPUT_PATH['report']
        self.find_relevant_paths(folder, extension)
        monotonic_reports, kind_reports, bisection_reports = self.extract_fields_for_three_strategies()
        return monotonic_reports, kind_reports, bisection_reports

    def extract_fields_as_list(self, reports: List[Stats], field):
        field_list = []
        i = 0
        for stat in reports:

            if field == TOTAL_RUNTIME:
                field_list.append(float(stat.total_runtime))
            elif field == N_SATS or field == N_UNSATS:
                field_list.append(int(stat.num_sats) + int(stat.num_unsats))
            else:
                print(f'ERROR!!! No such attribute in class Stats')
        return field_list

    def draw_scatter_plot_all(self, field: str):

        monotonic_reports, kind_reports, bisection_reports = self.collect_reports_for_three_strategies()


        folder, extension = OUTPUT_PATH['figure']
        os.makedirs(f'{folder}', exist_ok=True)
        filename1 = f'{folder}/{SCATTER}_{self.metric}_{MONOTONIC}_{KIND_BISECTION}.{extension}'
        filename2 = f'{folder}/{SCATTER}_{self.metric}_{MONOTONIC}_{BISECTION}.{extension}'

        # print(f'{monotonic_reports = }')
        monotonic_list = self.extract_fields_as_list(monotonic_reports, field)
        kind_list = self.extract_fields_as_list(kind_reports, field)
        bisection_list = self.extract_fields_as_list(bisection_reports, field)

        max_monotonic = max(monotonic_list)
        max_kind = max(kind_list)
        max_bisection = max(bisection_list)
        max_value1 = max(max_monotonic, max_kind)
        max_value2 = max(max_monotonic, max_bisection)



        if self.metric == WAE:
            color = 'blue'
        elif self.metric == WRE:
            color = 'red'
        elif self.metric == WHD:
            color = 'black'
        a = [1] * 10

        # monotonic_list = [1] * len(monotonic_list)

        plt.rc('font', size=SMALL_SIZE+6)  # controls default text sizes
        plt.rc('axes', titlesize=BIGGER_SIZE+6)  # fontsize of the axes title
        plt.rc('axes', labelsize=BIGGER_SIZE+5)  # fontsize of the x and y labels
        plt.rc('xtick', labelsize=MEDIUM_SIZE+1)  # fontsize of the tick labels
        plt.rc('ytick', labelsize=MEDIUM_SIZE+1)  # fontsize of the tick labels
        plt.rc('legend', fontsize=SMALL_SIZE)  # legend fontsize
        plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title
        plt.tight_layout()
        plt.rcParams["figure.figsize"] = (6, 5)
        plt.tight_layout()


        fig1, ax1 = plt.subplots()
        ax1.locator_params(tight=True, nbins=4)

        ErrorEval = 'ErrorEval'
        ax1.set( xlabel=f'bisection+ runtime', ylabel=f'{ErrorEval} runtime')
        ax1.set_title(f'{self.metric.upper()}: {ErrorEval} vs. bisection+')
        plt.scatter(kind_list, monotonic_list, c=color, s=10, marker='o', edgecolors=color, linewidths=1, alpha=0.1)
        plt.plot(list(np.arange(0, (max_value1), 0.01)), list(np.arange(0, (max_value1), 0.01)), c='grey')


        plt.xscale('log')
        plt.yscale('log')
        ax1.get_xaxis().set_major_formatter(matplotlib.ticker.LogFormatter())
        ax1.get_yaxis().set_major_formatter(matplotlib.ticker.LogFormatter())
        ax1.set_xticks([10 ** -2, 1, 100] * 10)
        ax1.set_yticks([10 ** -2, 1, 100] * 10)
        ax1.yaxis.set_label_coords(-.1, .5)
        plt.savefig(f'{filename1}')
        plt.close()

        # Figure2: Monotonic vs bisection
        fig2, ax2 = plt.subplots()


        ax2.set(xlabel=f'{BISECTION} runtime', ylabel=f'{ErrorEval} runtime')
        ax2.set_title(f'{self.metric.upper()}: {ErrorEval} vs. {BISECTION}')
        plt.scatter(bisection_list, monotonic_list, c=color, s=10, marker='o', edgecolors=color, linewidths=0.5,  alpha=0.1)
        plt.plot(list(np.arange(0, (max_value2), 0.01)), list(np.arange(0, (max_value2), 0.01)), c='grey')

        plt.yscale('log')
        plt.xscale('log')

        ax2.get_xaxis().set_major_formatter(matplotlib.ticker.LogFormatter())
        ax2.get_yaxis().set_major_formatter(matplotlib.ticker.LogFormatter())
        ax2.set_xticks([10**-2, 1, 100] * 10)
        ax2.set_yticks([10**-2, 1, 100] * 10)
        ax2.yaxis.set_label_coords(-.1, .5)
        plt.savefig(f'{filename2}')
        plt.close()

        pass

    def avg_runtime(self, field):
        monotonic_reports, kind_reports, bisection_reports = self.collect_reports_for_three_strategies()
        monotonic_list = self.extract_fields_as_list(monotonic_reports, field)
        kind_list = self.extract_fields_as_list(kind_reports, field)
        bisection_list = self.extract_fields_as_list(bisection_reports, field)

        total_count = len(monotonic_list)
        monotonic_is_less_than_kind = 0
        monotonic_is_less_than_bisection = 0

        for i in range(len(monotonic_list)):
            if monotonic_list[i] < kind_list[i]:
                monotonic_is_less_than_kind += 1
            if monotonic_list[i] < bisection_list[i]:
                monotonic_is_less_than_bisection += 1

        print(f'monotonic < bisection+ (%) = {(monotonic_is_less_than_kind/total_count) * 100}%')
        print(f'monotonic < bisection (%) = {(monotonic_is_less_than_bisection / total_count) * 100}%')

        print(f'{len(monotonic_list) = }')

    def draw_avg_plot_all(self, field):
        monotonic_reports, kind_reports, bisection_reports = self.collect_reports_for_three_strategies()
        folder, extension = OUTPUT_PATH['figure']
        os.makedirs(f'{folder}', exist_ok=True)
        filename1 = f'{folder}/{AVERAGE}_{self.metric}_{MONOTONIC}_{KIND_BISECTION}.{extension}'
        filename2 = f'{folder}/{AVERAGE}_{self.metric}_{MONOTONIC}_{BISECTION}.{extension}'


        monotonic_list = self.extract_fields_as_list(monotonic_reports, field)
        kind_list = self.extract_fields_as_list(kind_reports, field)
        bisection_list = self.extract_fields_as_list(bisection_reports, field)
        # print(monotonic_list[0])
        # print(kind_list[0])
        # print(bisection_list[0])


    # Create iterator object
    def __iter__(self):
        pass

    def __next__(self):
        pass

    def __repr__(self):
        return f'An object of class Result\n' \
               f'{self.benchmark = }\n' \
               f'{self.approximate_benchmark = }\n' \
               f'{self.experiment = }\n' \
               f'{self.strategy = }\n' \
               f'{self.metric = }\n'

    def __len__(self):
        return len(self.__dict__)



INPUT_PATH = {
    'ver': ('input/ver', 'v'),
    'gv': ('input/gv', 'gv'),
    'aig': ('input/aig', 'aig'),
    'dot': ('input/gv', 'dot'),
    'app_ver': ('input/ver', 'v'),
    'app_gv': ('input/gv', 'gv'),
    'report': ('output/report', 'csv')
}
OUTPUT_PATH = {
    'ver': ('output/ver', 'v'),
    'gv': ('output/gv', 'gv'),
    'aig': ('output/aig', 'aig'),
    'dot': ('output/gv', 'dot'),
    'z3': ('output/z3', 'py'),
    'report': ('output/report', 'csv'),
    'figure': ('output/figure', 'pdf')

}

TEST_PATH = {
    'tb': ('test/ver', 'v'),
    'iver': ('test/ver', 'iv'),
    'vvp': ('test/ver', 'txt'),
    'z3': ('test/z3', 'txt'),
}

LOG_PATH = {
    'iver': ('test/ver', 'log'),
    'z3': ('test/z3', 'log'),
    'yosys': ('output/log', 'log'),
    'lsoracle': ('output/log', 'log')
}




import numpy as np
import os
import copy
import csv
import shutil
from typing import *
import networkx as nx
from subprocess import PIPE
from .utils import *
from .graph import *


class Z3solver:
    def __init__(self, benchmark_name: str, approximate_benchmark_name: str = None, samples: list = [],
                 experiment: str = SINGLE,
                 pruned_percentage: int = None, pruned_gates=None, metric: str = WAE, precision: int = 4,
                 optimization: str = None, style: str = 'max'):
        """

        :param benchmark_name: the input benchmark in gv format
        :param approximate_benchmark_name: the approximate benchmark in gv format
        :param samples: number of samples for the mc evaluation; by defaults it is an empty list
        """
        self.__circuit_name = get_pure_name(benchmark_name)

        folder, extension = OUTPUT_PATH['gv']
        self.__graph_in_path = f'{folder}/{benchmark_name}.{extension}'

        self.__graph = Graph(benchmark_name, True)

        self.__pyscript_results_out_path = None

        folder, extension = LOG_PATH['z3']
        os.makedirs(f'{folder}/{benchmark_name}_{Z3}_{LOG}', exist_ok=True)
        self.__z3_log_path = f'{folder}/{benchmark_name}_{Z3}_{LOG}/{benchmark_name}_{Z3}_{LOG}.{extension}'

        self.__approximate_verilog_in_path = None
        self.__approximate_graph = None

        if approximate_benchmark_name:
            self.__approximate_circuit_name = approximate_benchmark_name
            folder, extension = INPUT_PATH['app_gv']
            self.__approximate_verilog_in_path = f'{folder}/{approximate_benchmark_name}.{extension}'
            self.__approximate_graph = Graph(approximate_benchmark_name, True)
            self.relabel_approximate_graph()
            self.approximate_graph.set_input_dict(self.approximate_graph.extract_inputs())
            self.approximate_graph.set_output_dict(self.approximate_graph.extract_outputs())
            self.approximate_graph.set_gate_dict(self.approximate_graph.extract_gates())
            self.approximate_graph.set_constant_dict(self.approximate_graph.extract_constants())

            self.__labeling_graph = copy.deepcopy(self.approximate_graph)

        self.__experiment = experiment
        self.__pruned_percentage = None
        # TODO
        # Later create and internal method that can generate pruned gates
        self.__pruned_gates = None
        if experiment == RANDOM:
            self.__pruned_percentage = pruned_percentage
            self.__pruned_gates = pruned_gates

        self.__metric = metric
        self.__precision = precision

        self.__z3_report = None

        self.__samples = samples
        self.__sample_results = None

        self.__z3string = None

        self.__z3pyscript = None

        self.__strategy = None

        self.__optimization = optimization

        self.__pyscript_files_for_labeling: list = []

        self.__z3_out_path = None

        self.__style = style


    @property
    def style(self):
        return self.__style

    @property
    def name(self):
        return self.__circuit_name

    @name.setter
    def name(self, this_name):
        self.__circuit_name = this_name

    @property
    def approximate_benchmark(self):
        return self.__approximate_circuit_name

    @property
    def graph_in_path(self):
        return self.__graph_in_path

    @property
    def out_path(self):
        return self.__z3_out_path

    def set_out_path(self, out_path: str):
        self.__z3_out_path = out_path

    @property
    def z3_log_path(self):
        return self.__z3_log_path

    @property
    def approximate_in_path(self):
        return self.__approximate_verilog_in_path

    @property
    def z3pyscript(self):
        return self.__z3pyscript

    @property
    def strategy(self):
        return self.__strategy

    def set_strategy(self, strategy: str):
        self.__strategy = strategy

    @property
    def optimization(self):
        return self.__optimization

    @optimization.setter
    def optimization(self, optimization):
        self.__optimization = optimization

    @property
    def metric(self):
        return self.__metric

    @metric.setter
    def metric(self, metric):
        self.__metric = metric

    @property
    def precision(self):
        return self.__precision

    @precision.setter
    def precision(self, precision):
        self.__precision = precision

    @property
    def experiment(self):
        return self.__experiment

    @experiment.setter
    def experiment(self, experiment):
        self.__experiment = experiment

    @property
    def pruned_percentage(self):
        return self.__pruned_percentage

    @pruned_percentage.setter
    def pruned_percentage(self, pruned_percentage):
        self.__pruned_percentage = pruned_percentage

    @property
    def pruned_gates(self):
        return self.__pruned_gates

    @pruned_gates.setter
    def pruned_gates(self, pruned_gates: List[int]):
        self.__pruned_gates = pruned_gates

    @property
    def z3_report(self):
        return self.__z3_report

    def set_z3_report(self, z3_report: str):
        self.__z3_report = z3_report

    @property
    def pyscript_files_for_labeling(self):
        return self.__pyscript_files_for_labeling

    def set_pyscript_files_for_labeling(self, pyscript_files_for_labeling):
        self.__pyscript_files_for_labeling = pyscript_files_for_labeling

    def append_pyscript_files_for_labeling(self, pyscript_file):
        self.__pyscript_files_for_labeling.append(pyscript_file)

    # TODO

    @property
    def z3string(self):
        return self.__z3string

    @property
    def samples(self):
        return self.__samples

    def set_samples(self, samples: np.array or list):
        self.__samples = samples

    @property
    def sample_results(self):
        return self.__sample_results

    def set_sample_results(self, results):
        self.__sample_results = results

    def import_results(self):
        arr = []
        with open(self.pyscript_results_out_path, 'r') as f:
            lines = f.readlines()
            for line in lines:
                line = line.replace('\n', '')
                arr.append(int(line))
        results = np.array(arr)
        return results

    def append_z3string(self, new_string):
        self.__z3string = f'{self.__z3string}\n' \
                          f'{new_string}'

    def set_z3string(self, new_string):
        self.__z3string = new_string

    @property
    def pyscript_results_out_path(self):
        return self.__pyscript_results_out_path

    def set_pyscript_results_out_path(self, pyscript_results_out_path: str):
        self.__pyscript_results_out_path = pyscript_results_out_path

    @property
    def graph(self):
        return self.__graph

    @property
    def approximate_graph(self):
        return self.__approximate_graph

    def set_approximate_graph(self, tmp_graph):
        self.__approximate_graph = tmp_graph

    @property
    def labeling_graph(self):
        return self.__labeling_graph

    def label_circuit(self, constant_value: bool = False):
        self.experiment = SINGLE
        self.set_strategy(MONOTONIC)
        removed_gate = []
        for key in self.labeling_graph.gate_dict:
            removed_gate = [self.labeling_graph.gate_dict[key]]
            self.create_pruned_z3pyscript_approximate(removed_gate, constant_value)
        self.run_z3pyscript_labeling()
        return self.import_labels(constant_value)

    def import_labels(self, constant_value: bool=False) -> Dict:
        label_dict: Dict[str, int] = {}
        folder, extension = OUTPUT_PATH['report']

        all_dirs = [f for f in os.listdir(folder)]
        # print(f'{all_dirs = }')
        relevant_dir = None
        for dir in all_dirs:
            if re.search(f'{self.approximate_benchmark}_labeling', dir) and os.path.isdir(f'{folder}/{dir}') and re.search(f'{constant_value}', dir):
                relevant_dir = f'{folder}/{dir}'
                # print(f'{relevant_dir = }')
        # print(f'{relevant_dir = }')
        all_csv = [f for f in os.listdir(relevant_dir)]
        for report in all_csv:
            if re.search(self.approximate_benchmark, report) and report.endswith(extension):
                gate_label = re.search('(g\d+)', report).group(1)

                with open(f'{relevant_dir}/{report}', 'r') as r:
                    csvreader = csv.reader(r)
                    for line in csvreader:
                        if re.search(WCE, line[0]):
                            gate_wce = int(line[1])

                            label_dict[gate_label] = gate_wce

        return label_dict




    # TODO
    # Deprecated
    def import_z3_expression(self):
        pass

    def set_z3pyscript(self, this_script):
        self.__z3pyscript = this_script

    def relabel_approximate_graph(self):
        gate_mapping = {}
        constant_mapping = {}
        output_mapping = {}
        for key in self.approximate_graph.gate_dict.keys():
            gate_mapping[self.approximate_graph.gate_dict[key]] = f'app_{self.approximate_graph.gate_dict[key]}'
        for key in self.approximate_graph.constant_dict.keys():
            constant_mapping[
                self.approximate_graph.constant_dict[key]] = f'app_{self.approximate_graph.constant_dict[key]}'
        for key in self.approximate_graph.output_dict.keys():
            output_mapping[self.approximate_graph.output_dict[key]] = f'app_{self.approximate_graph.output_dict[key]}'
        self.approximate_graph.set_graph(nx.relabel_nodes(self.approximate_graph.graph, gate_mapping))
        self.approximate_graph.set_graph(nx.relabel_nodes(self.approximate_graph.graph, constant_mapping))
        self.approximate_graph.set_graph(nx.relabel_nodes(self.approximate_graph.graph, output_mapping))

    def convert_gv_to_z3pyscript_test(self):
        folder, extension = OUTPUT_PATH['report']
        os.makedirs(f'{folder}/{self.name}', exist_ok=True)
        self.set_z3_report(f'{folder}/{self.name}/{self.name}_{TEST}.{extension}')

        folder, extension = OUTPUT_PATH['z3']
        os.makedirs(f'{folder}/{self.name}', exist_ok=True)
        self.set_out_path(f'{folder}/{self.name}/{self.name}_{TEST}.{extension}')

        folder, extension = TEST_PATH['z3']
        os.makedirs(f'{folder}/{self.name}', exist_ok=True)
        self.set_pyscript_results_out_path(f'{folder}/{self.name}/{self.name}_{TEST}.{extension}')

        import_string = self.create_imports()
        abs_function = self.create_abs_function()
        exact_circuit_declaration = self.declare_original_circuit()
        exact_circuit_expression = self.express_original_circuit()
        # TODO: Fix Later
        if self.metric == WHD:
            output_declaration = ''
            print(f'ERROR!!! Right now testing is not possible on WHD!')
        else:
            output_declaration = self.declare_original_output()
        exact_function = self.declare_original_function()
        solver = self.declare_solver()
        sample_expression = self.express_samples()
        store_results = self.store_results()
        self.set_z3pyscript(import_string + abs_function + exact_circuit_declaration + exact_circuit_expression +
                            output_declaration + exact_function + solver + sample_expression + store_results)

    def convert_gv_to_z3pyscript_maxerror_qor(self, strategy: str = DEFAULT_STRATEGY):

        self.experiment = QOR
        self.set_strategy(strategy)

        if self.metric == WRE:
            folder, extension = OUTPUT_PATH['report']
            if self.optimization == OPTIMIZE or self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                self.set_z3_report(
                    f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{self.optimization}.{extension}')
            else:
                self.set_z3_report(
                    f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}.{extension}')

            folder, extension = OUTPUT_PATH['z3']
            if self.optimization == OPTIMIZE or self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                self.set_out_path(
                    f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{self.optimization}.{extension}')
            else:
                self.set_out_path(
                    f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}.{extension}')

        else:
            folder, extension = OUTPUT_PATH['report']
            if self.optimization == OPTIMIZE or self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                self.set_z3_report(
                    f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_{self.strategy}_{self.optimization}.{extension}')
            else:
                self.set_z3_report(
                    f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_{self.strategy}.{extension}')

            folder, extension = OUTPUT_PATH['z3']
            if self.optimization == OPTIMIZE or self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                self.set_out_path(
                    f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_{self.strategy}_{self.optimization}.{extension}')
            else:
                self.set_out_path(
                    f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_{self.strategy}.{extension}')


        import_string = self.create_imports()
        abs_function = self.create_abs_function()

        # exact_part
        original_circuit_declaration = self.declare_original_circuit()
        original_circuit_expression = self.express_original_circuit()
        if self.metric == WHD:
            original_output_declaration = 'Blah Blah Blah\n'
        else:
            original_output_declaration = self.declare_original_output()

        # approximate_part
        approximate_circuit_declaration = self.declare_approximate_circuit()
        approximate_circuit_expression = self.express_approximate_circuit()
        if self.metric == WHD:
            approximate_output_declaration = 'Blah Blah Blah\n'
        else:
            approximate_output_declaration = self.declare_approximate_output()

        # error distance function
        declare_error_distance_function = self.declare_error_distance_function()
        # strategy

        strategy = self.express_strategy()

        self.set_z3pyscript(import_string + abs_function + original_circuit_declaration + original_circuit_expression +
                            original_output_declaration + approximate_circuit_declaration + approximate_circuit_expression +
                            approximate_output_declaration + declare_error_distance_function + strategy)

    def convert_gv_to_z3pyscript_maxerror_labeling(self, strategy: str = DEFAULT_STRATEGY):
        self.experiment = SINGLE
        self.set_strategy(strategy)
        removed_gate = []
        for key in self.graph.input_dict:
            removed_gate = [self.graph.input_dict[key]]
            self.create_pruned_z3pyscript(removed_gate)
        for key in self.graph.gate_dict:
            removed_gate = [self.graph.gate_dict[key]]
            self.create_pruned_z3pyscript(removed_gate)

    def convert_gv_to_z3pyscript_maxerror_random_pruning(self, strategy: str = DEFAULT_STRATEGY):

        self.experiment = RANDOM
        self.set_strategy(strategy)

        removed_gates = []
        for idx in self.pruned_gates:
            if idx < self.graph.num_inputs:  # remove pi
                removed_gates.append(self.graph.input_dict[idx])
            else:  # remove gate
                removed_gates.append(self.graph.gate_dict[idx - self.graph.num_inputs + 1])
        # print(f'{removed_gates}')
        self.create_pruned_z3pyscript(removed_gates)

    # TODO
    # Naming problems for more than one gate removal
    def create_pruned_z3pyscript(self, gates: list, constant_value: bool = False):
        self.create_pruned_graph_approximate(gates, constant_value)
        if self.experiment == SINGLE:
            gate = gates[0]
        # TODO
        elif self.experiment == RANDOM:
            gate = 'id0'
        folder, extension = OUTPUT_PATH['report']
        if self.metric == WRE:
            if self.optimization == OPTIMIZE or self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                folder = f'{folder}/{self.name}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{self.optimization}'
                os.makedirs(folder, exist_ok=True)
                self.set_z3_report(
                    f'{folder}/{self.name}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{self.optimization}_{gate}.{extension}')
            else:
                folder = f'{folder}/{self.name}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}'
                os.makedirs(folder, exist_ok=True)
                self.set_z3_report(
                    f'{folder}/{self.name}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{gate}.{extension}')
        else:

            if self.optimization == OPTIMIZE or self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                folder = f'{folder}/{self.name}_{self.experiment}_{self.metric}_{self.strategy}_{self.optimization}'
                os.makedirs(folder, exist_ok=True)
                self.set_z3_report(
                    f'{folder}/{self.name}_{self.experiment}_{self.metric}_{self.strategy}_{self.optimization}_{gate}.{extension}')
            else:
                folder = f'{folder}/{self.name}_{self.experiment}_{self.metric}_{self.strategy}'
                os.makedirs(folder, exist_ok=True)
                self.set_z3_report(
                    f'{folder}/{self.name}_{self.experiment}_{self.metric}_{self.strategy}_{gate}.{extension}')

        folder, extension = OUTPUT_PATH['z3']
        if self.metric == WRE:
            if self.optimization == OPTIMIZE or self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                folder = f'{folder}/{self.name}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{self.optimization}'
                os.makedirs(folder, exist_ok=True)
                self.set_out_path(
                    f'{folder}/{self.name}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{self.optimization}_{gate}.{extension}')
            else:
                folder = f'{folder}/{self.name}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}'
                os.makedirs(folder, exist_ok=True)
                self.set_out_path(
                    f'{folder}/{self.name}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{gate}.{extension}')
        else:
            if self.optimization == OPTIMIZE or self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                folder = f'{folder}/{self.name}_{self.experiment}_{self.metric}_{self.strategy}_{self.optimization}'
                os.makedirs(folder, exist_ok=True)
                self.set_out_path(
                    f'{folder}/{self.name}_{self.experiment}_{self.metric}_{self.strategy}_{self.optimization}_{gate}.{extension}')
            else:
                folder = f'{folder}/{self.name}_{self.experiment}_{self.metric}_{self.strategy}'
                os.makedirs(folder, exist_ok=True)
                self.set_out_path(
                    f'{folder}/{self.name}_{self.experiment}_{self.metric}_{self.strategy}_{gate}.{extension}')

        self.append_pyscript_files_for_labeling(self.out_path)

        import_string = self.create_imports()
        abs_function = self.create_abs_function()

        # exact_part
        original_circuit_declaration = self.declare_original_circuit()
        original_circuit_expression = self.express_original_circuit()
        if self.metric == WHD:
            original_output_declaration = f'\n'
        else:
            original_output_declaration = self.declare_original_output()

        approximate_circuit_declaration = self.declare_approximate_circuit()
        approximate_circuit_expression = self.express_approximate_circuit()

        if self.metric == WHD:
            xor_miter_declaration = self.declare_xor_miter()
        else:
            approximate_output_declaration = self.declare_approximate_output()

        # error distance function
        declare_error_distance_function = self.declare_error_distance_function()
        # strategy

        strategy = self.express_strategy()

        if self.metric == WHD:
            self.set_z3pyscript(
                import_string + abs_function + original_circuit_declaration + original_circuit_expression +
                approximate_circuit_expression + xor_miter_declaration + declare_error_distance_function + strategy)
        else:
            self.set_z3pyscript(
                import_string + abs_function + original_circuit_declaration + original_circuit_expression +
                original_output_declaration + approximate_circuit_declaration + approximate_circuit_expression +
                approximate_output_declaration + declare_error_distance_function + strategy)

        self.export_z3pyscript()

    def create_pruned_z3pyscript_approximate(self, gates: list, constant_value: bool = False):
        self.create_pruned_graph_approximate(gates, constant_value)
        if self.experiment == SINGLE:
            gate = gates[0]
        # TODO
        elif self.experiment == RANDOM:
            gate = 'id0'
        folder, extension = OUTPUT_PATH['report']
        if self.metric == WRE:
            if self.optimization == OPTIMIZE or self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                folder = f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{self.optimization}_{constant_value}'
                os.makedirs(folder, exist_ok=True)
                self.set_z3_report(
                    f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{self.optimization}_{gate}.{extension}')
            else:
                folder = f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{constant_value}'
                os.makedirs(folder, exist_ok=True)
                self.set_z3_report(
                    f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{gate}.{extension}')
        else:

            if self.optimization == OPTIMIZE or self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                folder = f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_{self.strategy}_{self.optimization}_{constant_value}'
                os.makedirs(folder, exist_ok=True)
                self.set_z3_report(
                    f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_{self.strategy}_{self.optimization}_{gate}.{extension}')
            else:
                folder = f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_{self.strategy}_{constant_value}'
                os.makedirs(folder, exist_ok=True)
                self.set_z3_report(
                    f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_{self.strategy}_{gate}.{extension}')

        folder, extension = OUTPUT_PATH['z3']
        if self.metric == WRE:
            if self.optimization == OPTIMIZE or self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                folder = f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{self.optimization}_{constant_value}'
                os.makedirs(folder, exist_ok=True)
                self.set_out_path(
                    f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{self.optimization}_{gate}.{extension}')
            else:
                folder = f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{constant_value}'
                os.makedirs(folder, exist_ok=True)
                self.set_out_path(
                    f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_d{self.precision}_{self.strategy}_{gate}.{extension}')
        else:
            if self.optimization == OPTIMIZE or self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                folder = f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_{self.strategy}_{self.optimization}_{constant_value}'
                os.makedirs(folder, exist_ok=True)
                self.set_out_path(
                    f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_{self.strategy}_{self.optimization}_{gate}.{extension}')
            else:
                folder = f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_{self.strategy}_{constant_value}'
                os.makedirs(folder, exist_ok=True)
                self.set_out_path(
                    f'{folder}/{self.approximate_benchmark}_{self.experiment}_{self.metric}_{self.strategy}_{gate}.{extension}')

        self.append_pyscript_files_for_labeling(self.out_path)

        import_string = self.create_imports()
        abs_function = self.create_abs_function()

        # exact_part
        original_circuit_declaration = self.declare_original_circuit()
        original_circuit_expression = self.express_original_circuit()
        if self.metric == WHD:
            original_output_declaration = f'\n'
        else:
            original_output_declaration = self.declare_original_output()

        approximate_circuit_declaration = self.declare_approximate_circuit()
        approximate_circuit_expression = self.express_approximate_circuit()

        if self.metric == WHD:
            xor_miter_declaration = self.declare_xor_miter()
        else:
            approximate_output_declaration = self.declare_approximate_output()

        # error distance function
        declare_error_distance_function = self.declare_error_distance_function()
        # strategy

        strategy = self.express_strategy()

        if self.metric == WHD:
            self.set_z3pyscript(
                import_string + abs_function + original_circuit_declaration + original_circuit_expression +
                approximate_circuit_expression + xor_miter_declaration + declare_error_distance_function + strategy)
        else:
            self.set_z3pyscript(
                import_string + abs_function + original_circuit_declaration + original_circuit_expression +
                original_output_declaration + approximate_circuit_declaration + approximate_circuit_expression +
                approximate_output_declaration + declare_error_distance_function + strategy)

        self.export_z3pyscript()

    def convert_gv_to_z3pyscript_xpat(self):
        pass

    def create_pruned_graph(self, gates: list):

        tmp_graph = copy.deepcopy(self.graph)
        self.set_approximate_graph(tmp_graph)
        mapping_dict = {}
        for gate in gates:
            if self.graph.is_pi(gate) or self.graph.is_cleaned_pi(gate) or self.graph.is_pruned_pi(gate):
                mapping_dict[gate] = f'app_{gate}'
        for gate in gates:
            self.approximate_graph.graph.nodes[gate][PRUNED] = True
        self.approximate_graph.set_graph(nx.relabel_nodes(self.approximate_graph.graph, mapping_dict))
        self.relabel_approximate_graph()
        self.approximate_graph.set_input_dict(self.approximate_graph.extract_inputs())
        self.approximate_graph.set_output_dict(self.approximate_graph.extract_outputs())
        self.approximate_graph.set_gate_dict(self.approximate_graph.extract_gates())
        self.approximate_graph.set_constant_dict(self.approximate_graph.extract_constants())

    def create_pruned_graph_approximate(self, gates: list, constant_value: bool =False):

        tmp_graph = copy.deepcopy(self.labeling_graph)
        self.set_approximate_graph(tmp_graph)
        mapping_dict = {}
        for gate in gates:
            if self.labeling_graph.is_pi(gate) or self.labeling_graph.is_cleaned_pi(gate) or self.labeling_graph.is_pruned_pi(gate):
                mapping_dict[gate] = f'{gate}'
        for gate in gates:
            self.approximate_graph.graph.nodes[gate][PRUNED] =  constant_value
        self.approximate_graph.set_graph(nx.relabel_nodes(self.approximate_graph.graph, mapping_dict))
        self.relabel_approximate_graph()
        self.approximate_graph.set_input_dict(self.approximate_graph.extract_inputs())
        self.approximate_graph.set_output_dict(self.approximate_graph.extract_outputs())
        self.approximate_graph.set_gate_dict(self.approximate_graph.extract_gates())
        self.approximate_graph.set_constant_dict(self.approximate_graph.extract_constants())

    # TODO
    # for other back-ends as well

    def declare_error_distance_function(self):
        ed_function = ''

        if self.metric == WAE:
            ed_function += f"f_exact = Function('f_exact', IntSort(), IntSort())\n" \
                           f"f_approx = Function('f_approx', IntSort(), IntSort())\n"
            ed_function += f"f_error = Function('f_error', IntSort(), IntSort(), IntSort())\n"
        elif self.metric == WHD:
            ed_function += f"f_error = Function('f_error', IntSort(), "
            for i in range(self.graph.num_outputs):
                if i == self.graph.num_outputs - 1:
                    ed_function += f"IntSort() )\n"
                else:
                    ed_function += f"IntSort(), "
        elif self.metric == WRE:
            ed_function += f"f_exact = Function('f_exact', IntSort(), IntSort())\n" \
                           f"f_approx = Function('f_approx', IntSort(), IntSort())\n"
            ed_function += f"f_error = Function('f_error', IntSort(), IntSort(), RealSort())\n"
        ed_function += f'\n'
        return ed_function

    def express_strategy(self, metric: str = None):
        strategy_expressed = ''
        if re.search(MONOTONIC, self.strategy):
            strategy_expressed += self.express_monotonic_strategy()
        elif re.search(BISECTION, self.strategy):
            strategy_expressed += self.express_bisection_strategy()
        elif re.search(MC, self.strategy):
            strategy_expressed += self.express_mc_strategy()
        elif re.search(KIND_BISECTION, self.strategy):
            strategy_expressed += self.express_kind_bisection_strategy()
        else:
            print(f'ERROR!!! no strategy is specified!')
            exit()
        return strategy_expressed

    def declare_stats(self):
        stats = ''
        stats = f'foundWCE = False\n' \
                f'stats: dict = {{}}\n' \
                f"stats['wce'] = 0\n"

        if self.metric == WAE:
            stats += f"stats['et'] = 0\n"
        elif self.metric == WHD:
            stats += f"stats['et'] = 0\n"
        elif self.metric == WRE:
            stats += f"stats['et'] = float(\"{{:.{self.precision}f}}\".format(0))\n"

        stats += f"stats['num_sats'] = 0\n" \
                 f"stats['num_unsats'] = 0\n" \
                 f"stats['sat_runtime'] = 0.0\n" \
                 f"stats['unsat_runtime'] = 0.0\n" \
                 f"stats['jumps'] = []\n"
        if self.metric == WAE or self.metric == WRE:
            stats += f"max = (2 ** {self.graph.num_outputs}) -1\n"
        elif self.metric == WHD:
            stats += f"max = {self.graph.num_outputs}\n"
        return stats

    def express_mc_while_loop(self):
        loop = ''
        loop += f's=Solver()\n' \
                f'start_whole = time.time()\n'

        for sample in self.samples:
            loop += f's.push()\n' \
                    f's.add(f_error(exact_out, approx_out) == z3_abs(exact_out - approx_out))\n' \
                    f'{self.express_push_input_sample(sample)}' \
                    f'response=s.check()\n' \
                    f'returned_model = s.model()\n' \
                    f'returned_value = abs(int(returned_model[f_error].else_value().as_long()))\n' \
                    f'returned_value_reval = abs(int(returned_model.evaluate(f_error(exact_out, approx_out)).as_long()))\n' \
                    f'if returned_value == returned_value_reval:\n' \
                    f"{TAB}print(f'double-check is passed!')\n" \
                    f"else:\n" \
                    f"{TAB}print(f'ERROR!!! double-check failed! exiting...')\n" \
                    f"{TAB}exit()\n" \
                    f"if returned_value > stats['wce']:\n" \
                    f"{TAB}stats['wce'] = returned_value\n" \
                    f"s.pop()\n" \
                    f"\n" \
                    f""
        loop += self.express_stats()
        return loop

    def express_push_input_sample(self, sample: int):
        sample_expression = ''
        s_expression = [True if i == '1' else False for i in list(f'{sample:0{self.graph.num_inputs}b}')]
        s_expression.reverse()
        sample_expression += f's.add('
        for idx, e in enumerate(s_expression):
            if idx == len(s_expression) - 1:
                sample_expression += f'{self.graph.input_dict[idx]}=={e})\n'
            else:
                sample_expression += f'{self.graph.input_dict[idx]}=={e}, '

        return sample_expression

    def express_kind_bisection_while_loop(self):
        loop = ''
        if self.metric == WAE or self.metric == WRE:
            loop += f'upper_bound = 2**({self.graph.num_outputs}) - 1\n' \
                    f'lower_bound = 0 \n' \
                    f'start_whole = time.time()\n'
        else:
            loop += f'upper_bound = {self.graph.num_outputs}\n' \
                    f'lower_bound = 0 \n' \
                    f'start_whole = time.time()\n'

        loop += f's = Solver()\n'

        loop += f'while(not foundWCE):\n'

        if self.metric == WAE:
            loop += f"{TAB}stats['et'] = (upper_bound + lower_bound) // 2\n"
        elif self.metric == WHD:
            loop += f"{TAB}stats['et'] = (upper_bound + lower_bound) // 2\n"
        elif self.metric == WRE:
            loop += f"{TAB}import math\n"
            loop += f"{TAB}stats['et'] = math.floor((float(upper_bound + lower_bound) / 2) * {10 ** self.precision}) / {10 ** self.precision}\n"



        # Check termination
        if self.metric == WAE:
            loop += f"{TAB}if upper_bound - lower_bound <= 1:\n" \
                    f"{TAB}{TAB}foundWCE = True\n" \
                    f"{TAB}{TAB}stats['wce'] = lower_bound\n" \
                    f"{TAB}{TAB}if stats['et'] in stats['jumps']:\n" \
                    f"{TAB}{TAB}{TAB}break\n"
        elif self.metric == WHD:
            loop += f"{TAB}if upper_bound - lower_bound <= 1:\n" \
                    f"{TAB}{TAB}foundWCE = True\n" \
                    f"{TAB}{TAB}if lower_bound == 0:\n" \
                    f"{TAB}{TAB}{TAB}stats['wce'] = lower_bound\n" \
                    f"{TAB}{TAB}else:\n" \
                    f"{TAB}{TAB}{TAB}stats['wce'] = upper_bound\n" \
                    f"{TAB}{TAB}if stats['et'] in stats['jumps']:\n" \
                    f"{TAB}{TAB}{TAB}break\n"
        elif self.metric == WRE:
            loop += f"{TAB}if stats['et'] == lower_bound:\n" \
                    f"{TAB}{TAB}if stats['et'] in stats['jumps']:\n" \
                    f"{TAB}{TAB}{TAB}if lower_bound == 0:\n" \
                    f"{TAB}{TAB}{TAB}{TAB}foundWCE = True\n" \
                    f"{TAB}{TAB}{TAB}{TAB}stats['wce'] = lower_bound\n" \
                    f"{TAB}{TAB}{TAB}else:\n" \
                    f"{TAB}{TAB}{TAB}{TAB}foundWCE = True\n" \
                    f"{TAB}{TAB}{TAB}{TAB}stats['wce'] = upper_bound\n"

            loop += f"{TAB}elif round(upper_bound - lower_bound, 2) <= (10 ** -{self.precision}):\n" \
                    f"{TAB}{TAB}foundWCE = True\n" \
                    f"{TAB}{TAB}stats['wce'] = lower_bound\n" \
                    f"{TAB}{TAB}if stats['et'] in stats['jumps']:\n" \
                    f"{TAB}{TAB}{TAB}break\n"
        loop += f"{TAB}if stats['et'] not in stats['jumps']:\n" \
                f"{TAB}{TAB}stats['jumps'].append(stats['et'])\n"

        loop += f'{TAB}start_iteration = time.time()\n' \
                f'{TAB}s.push()\n'
        if self.metric == WAE or self.metric == WRE:
            loop += f'{TAB}s.add(f_exact(exact_out) == exact_out)\n' \
                    f'{TAB}s.add(f_approx(approx_out) == approx_out)\n'

        if self.metric == WAE:
            loop += f'{TAB}s.add(f_error(exact_out, approx_out) == exact_out - approx_out)\n' \
                    f"{TAB}s.add(z3_abs(f_error(exact_out, approx_out)) > stats['et'])\n"
        elif self.metric == WHD:
            if self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                pass
            else:
                loop += f"{TAB}s.add(f_error("
                for i in range(self.graph.num_outputs):
                    if i == self.graph.num_outputs - 1:
                        loop += f"o{i}_{XOR}_{INT}) == "
                    else:
                        loop += f"o{i}_{XOR}_{INT}, "
                for i in range(self.graph.num_outputs):
                    if i == self.graph.num_outputs - 1:
                        loop += f"o{i}_{XOR}_{INT})\n"
                    else:
                        loop += f"o{i}_{XOR}_{INT} +  "
                loop += f"{TAB}s.add(f_error("
                for i in range(self.graph.num_outputs):
                    if i == self.graph.num_outputs - 1:
                        loop += f"o{i}_{XOR}_{INT}) > stats['et'])\n"
                    else:
                        loop += f"o{i}_{XOR}_{INT}, "
            # TODO
        elif self.metric == WRE:
            loop += f'{TAB}s.add(f_error(exact_out, approx_out) == z3_abs(exact_out - approx_out) / (z3_abs(exact_out) + z3_abs(1.0))  )\n' \
                    f"{TAB}s.add(z3_abs(f_error(exact_out, approx_out)) > stats['et'])\n"

        loop += f"{TAB}response = s.check()\n"
        loop += self.express_kind_bisection_while_loop_sat()
        loop += self.express_kind_bisection_while_loop_unsat()
        loop += self.express_stats()

        return loop

    def express_bisection_while_loop(self):
        loop = ''
        if self.metric == WAE or self.metric == WRE:
            loop += f'upper_bound = 2**({self.graph.num_outputs}) - 1\n' \
                    f'lower_bound = 0 \n' \
                    f'start_whole = time.time()\n'
        else:
            loop += f'upper_bound = {self.graph.num_outputs}\n' \
                    f'lower_bound = 0 \n' \
                    f'start_whole = time.time()\n'

        loop += f's = Solver()\n'

        loop += f'while(not foundWCE):\n'


        if self.metric == WAE:
            loop += f"{TAB}stats['et'] = (upper_bound + lower_bound) // 2\n"
        elif self.metric == WHD:
            loop += f"{TAB}stats['et'] = (upper_bound + lower_bound) // 2\n"
        elif self.metric == WRE:
            loop += f"{TAB}import math\n"
            loop += f"{TAB}stats['et'] = math.floor((float(upper_bound + lower_bound) / 2) * {10 ** self.precision}) / {10 ** self.precision}\n"



        # Check termination
        if self.metric == WAE:
            loop += f"{TAB}if upper_bound - lower_bound <= 1:\n" \
                    f"{TAB}{TAB}foundWCE = True\n" \
                    f"{TAB}{TAB}if lower_bound == 0:\n" \
                    f"{TAB}{TAB}{TAB}stats['wce'] = lower_bound\n" \
                    f"{TAB}{TAB}else:\n" \
                    f"{TAB}{TAB}{TAB}stats['wce'] = upper_bound\n" \
                    f"{TAB}{TAB}if stats['et'] in stats['jumps']:\n" \
                    f"{TAB}{TAB}{TAB}break\n"
        elif self.metric == WHD:
            loop += f"{TAB}if upper_bound - lower_bound <= 1:\n" \
                    f"{TAB}{TAB}foundWCE = True\n" \
                    f"{TAB}{TAB}if lower_bound == 0:\n" \
                    f"{TAB}{TAB}{TAB}stats['wce'] = lower_bound\n" \
                    f"{TAB}{TAB}else:\n" \
                    f"{TAB}{TAB}{TAB}stats['wce'] = upper_bound\n" \
                    f"{TAB}{TAB}if stats['et'] in stats['jumps']:\n" \
                    f"{TAB}{TAB}{TAB}break\n"
        elif self.metric == WRE:
            loop += f"{TAB}if stats['et'] == lower_bound:\n" \
                    f"{TAB}{TAB}if stats['et'] in stats['jumps']:\n" \
                    f"{TAB}{TAB}{TAB}if lower_bound == 0:\n" \
                    f"{TAB}{TAB}{TAB}{TAB}foundWCE = True\n" \
                    f"{TAB}{TAB}{TAB}{TAB}stats['wce'] = lower_bound\n" \
                    f"{TAB}{TAB}{TAB}else:\n" \
                    f"{TAB}{TAB}{TAB}{TAB}foundWCE = True\n" \
                    f"{TAB}{TAB}{TAB}{TAB}stats['wce'] = upper_bound\n"
            loop += f"{TAB}if round(upper_bound - lower_bound, 2) <= (10 ** -{self.precision}):\n" \
                    f"{TAB}{TAB}foundWCE = True\n" \
                    f"{TAB}{TAB}if lower_bound == 0:\n" \
                    f"{TAB}{TAB}{TAB}stats['wce'] = lower_bound\n" \
                    f"{TAB}{TAB}else:\n" \
                    f"{TAB}{TAB}{TAB}stats['wce'] = upper_bound\n" \
                    f"{TAB}{TAB}if stats['et'] in stats['jumps']:\n" \
                    f"{TAB}{TAB}{TAB}break\n"

        loop += f"{TAB}if stats['et'] not in stats['jumps']:\n" \
                f"{TAB}{TAB}stats['jumps'].append(stats['et'])\n"

        loop += f'{TAB}start_iteration = time.time()\n' \
                f'{TAB}s.push()\n'
        if self.metric == WAE or self.metric == WRE:
                loop += f'{TAB}s.add(f_exact(exact_out) == exact_out)\n' \
                        f'{TAB}s.add(f_approx(approx_out) == approx_out)\n'


        if self.metric == WAE:
            loop += f'{TAB}s.add(f_error(exact_out, approx_out) == exact_out - approx_out)\n' \
                    f"{TAB}s.add(z3_abs(f_error(exact_out, approx_out)) > stats['et'])\n"
        elif self.metric == WHD:
            if self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                pass
            else:
                loop += f"{TAB}s.add(f_error("
                for i in range(self.graph.num_outputs):
                    if i == self.graph.num_outputs - 1:
                        loop += f"o{i}_{XOR}_{INT}) == "
                    else:
                        loop += f"o{i}_{XOR}_{INT}, "
                for i in range(self.graph.num_outputs):
                    if i == self.graph.num_outputs - 1:
                        loop += f"o{i}_{XOR}_{INT})\n"
                    else:
                        loop += f"o{i}_{XOR}_{INT} +  "
                loop += f"{TAB}s.add(f_error("
                for i in range(self.graph.num_outputs):
                    if i == self.graph.num_outputs - 1:
                        loop += f"o{i}_{XOR}_{INT}) > stats['et'])\n"
                    else:
                        loop += f"o{i}_{XOR}_{INT}, "
            # TODO
        elif self.metric == WRE:
            loop += f'{TAB}s.add(f_error(exact_out, approx_out) == z3_abs(exact_out - approx_out) / (z3_abs(exact_out) + z3_abs(1.0))  )\n' \
                    f"{TAB}s.add(z3_abs(f_error(exact_out, approx_out)) > stats['et'])\n"

        loop += f"{TAB}response = s.check()\n"

        loop += self.express_bisection_while_loop_sat()
        loop += self.express_bisection_while_loop_unsat()
        loop += self.express_stats()

        return loop

    def express_monotonic_while_loop(self):
        loop = ''

        loop += f'start_whole = time.time()\n'
        # print(f'{self.optimization == MAXIMIZE = }')
        # print(f'{self.optimization = }')
        if self.optimization == OPTIMIZE or self.optimization == MAXIMIZE and (self.strategy != BISECTION):
            loop += f's = Optimize()\n'
        else:
            # print('We are here')
            loop += f's = Solver()\n'

        loop += f"stats['jumps'].append(stats['et'])\n" \
                f'while(not foundWCE):\n' \
                f'{TAB}start_iteration = time.time()\n' \
                f'{TAB}s.push()\n'
        if self.metric == WAE or self.metric == WRE:
            loop += f'{TAB}s.add(f_exact(exact_out) == exact_out)\n' \
                    f'{TAB}s.add(f_approx(approx_out) == approx_out)\n'

        if self.metric == WAE:
            if self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                loop += f'{TAB}s.add(f_error(exact_out, approx_out) == exact_out - approx_out)\n'
                if self.style == 'min':
                    loop += f"{TAB}s.add(z3_abs(f_error(exact_out, approx_out)) <= max)\n" \
                            f"{TAB}s.add(z3_abs(f_error(exact_out, approx_out)) > 0)\n" \
                            f"{TAB}s.minimize(z3_abs(f_error(exact_out, approx_out)))\n"
                elif self.style == 'max':
                    loop += f"{TAB}s.add(z3_abs(f_error(exact_out, approx_out)) > stats['et'])\n" \
                            f"{TAB}s.maximize(z3_abs(f_error(exact_out, approx_out)))\n"
            else:
                loop += f'{TAB}s.add(f_error(exact_out, approx_out) == exact_out - approx_out)\n' \
                        f"{TAB}s.add(z3_abs(f_error(exact_out, approx_out)) > stats['et'])\n"

            # TODO add optimization thingy right here
        elif self.metric == WHD:
            if self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                pass
            else:
                loop += f"{TAB}s.add(f_error("
                for i in range(self.graph.num_outputs):
                    if i == self.graph.num_outputs - 1:
                        loop += f"o{i}_{XOR}_{INT}) == "
                    else:
                        loop += f"o{i}_{XOR}_{INT}, "
                for i in range(self.graph.num_outputs):
                    if i == self.graph.num_outputs - 1:
                        loop += f"o{i}_{XOR}_{INT})\n"
                    else:
                        loop += f"o{i}_{XOR}_{INT} +  "
                loop += f"{TAB}s.add(f_error("
                for i in range(self.graph.num_outputs):
                    if i == self.graph.num_outputs - 1:
                        loop += f"o{i}_{XOR}_{INT}) > stats['et'])\n"
                    else:
                        loop += f"o{i}_{XOR}_{INT}, "
            # TODO
        elif self.metric == WRE:
            if self.optimization == MAXIMIZE and (self.strategy != BISECTION):
                loop += f'{TAB}s.add(f_error(exact_out, approx_out) == z3_abs(exact_out - approx_out) / (z3_abs(exact_out) + z3_abs(1.0))  )\n' \
                        f"{TAB}s.add(z3_abs(f_error(exact_out, approx_out)) > stats['et'])\n" \
                        f"{TAB}s.maximize(z3_abs(f_error(exact_out, approx_out)))\n"
            else:
                loop += f'{TAB}s.add(f_error(exact_out, approx_out) == z3_abs(exact_out - approx_out) / (z3_abs(exact_out) + z3_abs(1.0))  )\n' \
                        f"{TAB}s.add(z3_abs(f_error(exact_out, approx_out)) > stats['et'])\n"
        loop += f"{TAB}response = s.check()\n"

        loop += self.express_monotonic_while_loop_sat()
        loop += self.express_monotonic_while_loop_unsat()
        loop += self.express_stats()

        return loop

    def express_kind_bisection_while_loop_sat(self):
        if_sat = ''

        if_sat += f"{TAB}if response == sat:\n" \
                  f"{TAB}{TAB}print(f'sat')\n" \
                  f"{TAB}{TAB}end_iteration = time.time()\n" \
                  f"{TAB}{TAB}returned_model = s.model()\n" \
                  f"{TAB}{TAB}print(f\"{{returned_model[f_error].else_value() = }}\")\n"

        if self.metric == WAE:
            if_sat += f"{TAB}{TAB}returned_value = abs(int(returned_model[f_error].else_value().as_long()))\n" \
                      f"{TAB}{TAB}returned_value_reval = abs(int(returned_model.evaluate(f_error(exact_out, approx_out)).as_long()))\n"
        elif self.metric == WHD:
            if_sat += f"{TAB}{TAB}returned_value = abs(int(returned_model[f_error].else_value().as_long()))\n"

            if_sat += f"{TAB}{TAB}returned_value_reval = abs(int(returned_model.evaluate(f_error("
            for i in range(self.graph.num_outputs):
                if i == self.graph.num_outputs - 1:
                    if_sat += f"o{i}_{XOR}_{INT})).as_long()))\n"
                else:
                    if_sat += f"o{i}_{XOR}_{INT}, "
            # TODO
        elif self.metric == WRE:
            if_sat += f"{TAB}{TAB}returned_value = ((returned_model[f_error].else_value().as_decimal({self.precision})))\n" \
                      f"{TAB}{TAB}returned_value_reval = ((returned_model.evaluate(f_error(exact_out, approx_out)).as_decimal({self.precision})))\n"


        # Double-check
        if_sat += f"{TAB}{TAB}if returned_value == returned_value_reval:\n" \
                  f"{TAB}{TAB}{TAB}print(f'double-check is passed!')\n" \
                  f"{TAB}{TAB}else:\n" \
                  f"{TAB}{TAB}{TAB}print(f'ERROR!!! double-check failed! exiting...')\n" \
                  f"{TAB}{TAB}{TAB}exit()\n"

        if self.metric == WRE:
            if_sat += f"{TAB}{TAB}if returned_value[-1] == '?':\n" \
                      f"{TAB}{TAB}{TAB}print('removing the last question mark!')\n" \
                      f"{TAB}{TAB}{TAB}returned_value = abs(float(returned_value[:-1])) + 10 ** -({self.precision})\n" \
                      f"{TAB}{TAB}else:\n" \
                      f"{TAB}{TAB}{TAB}returned_value = abs(float(returned_value))\n"

        if self.metric == WAE:
            if_sat += f"{TAB}{TAB}if upper_bound - lower_bound <= 1:\n" \
                      f"{TAB}{TAB}{TAB}foundWCE = True\n" \
                      f"{TAB}{TAB}{TAB}stats['wce'] = returned_value\n" \
                      f"{TAB}{TAB}else:\n" \
                      f"{TAB}{TAB}{TAB}lower_bound = returned_value\n"
        elif self.metric == WHD:
            if_sat += f"{TAB}{TAB}if upper_bound - lower_bound <= 1:\n" \
                      f"{TAB}{TAB}{TAB}foundWCE = True\n" \
                      f"{TAB}{TAB}{TAB}stats['wce'] = returned_value\n" \
                      f"{TAB}{TAB}else:\n" \
                      f"{TAB}{TAB}{TAB}lower_bound = returned_value\n"
        elif self.metric == WRE:
            if_sat += f"{TAB}{TAB}if round(upper_bound - lower_bound, 2) <= (10 ** -{self.precision}):\n" \
                      f"{TAB}{TAB}{TAB}foundWCE = True\n" \
                      f"{TAB}{TAB}{TAB}stats['wce'] = float(\"{{:.{self.precision}f}}\".format(returned_value))\n" \
                      f"{TAB}{TAB}else:\n" \
                      f"{TAB}{TAB}{TAB}lower_bound = float(\"{{:.{self.precision}f}}\".format(returned_value))\n"

        if_sat += f"{TAB}{TAB}stats['num_sats'] += 1\n" \
                  f"{TAB}{TAB}stats['sat_runtime'] += (end_iteration - start_iteration)\n" \
                  f"{TAB}{TAB}if stats['et'] == max:\n" \
                  f"{TAB}{TAB}{TAB}foundWCE = True\n" \
                  f"{TAB}{TAB}{TAB}stats['wce'] = stats['et']\n"

        return if_sat

    def express_bisection_while_loop_sat(self):
        if_sat = ''

        if_sat += f"{TAB}if response == sat:\n" \
                  f"{TAB}{TAB}print(f'sat')\n" \
                  f"{TAB}{TAB}end_iteration = time.time()\n" \
                  f"{TAB}{TAB}returned_model = s.model()\n" \
                  f"{TAB}{TAB}print(f'{{returned_model = }}')\n"
        if self.metric == WAE or self.metric == WRE:
            f"{TAB}{TAB}print(f\"{{returned_model[f_exact].else_value() = }}\")\n" \
            f"{TAB}{TAB}print(f\"{{returned_model[f_approx].else_value() = }}\")\n" \
            f"{TAB}{TAB}print(f\"{{returned_model[f_error].else_value() = }}\")\n"
        elif self.metric == WHD:
            f"{TAB}{TAB}print(f\"{{returned_model[f_error].else_value() = }}\")\n"


        if self.metric == WAE:
            if_sat += f"{TAB}{TAB}returned_value = abs(int(returned_model[f_error].else_value().as_long()))\n" \
                      f"{TAB}{TAB}returned_value_reval = abs(int(returned_model.evaluate(f_error(exact_out, approx_out)).as_long()))\n"
        elif self.metric == WHD:
            if_sat += f"{TAB}{TAB}returned_value = abs(int(returned_model[f_error].else_value().as_long()))\n"

            if_sat += f"{TAB}{TAB}returned_value_reval = abs(int(returned_model.evaluate(f_error("
            for i in range(self.graph.num_outputs):
                if i == self.graph.num_outputs - 1:
                    if_sat += f"o{i}_{XOR}_{INT})).as_long()))\n"
                else:
                    if_sat += f"o{i}_{XOR}_{INT}, "
            # TODO
        elif self.metric == WRE:
            if_sat += f"{TAB}{TAB}returned_value = ((returned_model[f_error].else_value().as_decimal({self.precision})))\n" \
                      f"{TAB}{TAB}returned_value_reval = ((returned_model.evaluate(f_error(exact_out, approx_out)).as_decimal({self.precision})))\n"

        if_sat += f"{TAB}{TAB}if returned_value == returned_value_reval:\n" \
                  f"{TAB}{TAB}{TAB}print(f'double-check is passed!')\n" \
                  f"{TAB}{TAB}else:\n" \
                  f"{TAB}{TAB}{TAB}print(f'ERROR!!! double-check failed! exiting...')\n" \
                  f"{TAB}{TAB}{TAB}exit()\n"

        if self.metric == WRE:
            if_sat += f"{TAB}{TAB}if returned_value[-1] == '?':\n" \
                      f"{TAB}{TAB}{TAB}print('removing the last question mark!')\n" \
                      f"{TAB}{TAB}{TAB}returned_value = abs(float(returned_value[:-1])) + 10 ** -(2)\n" \
                      f"{TAB}{TAB}else:\n" \
                      f"{TAB}{TAB}{TAB}returned_value = abs(float(returned_value))\n"


        if self.metric == WAE:
            if_sat += f"{TAB}{TAB}if upper_bound - lower_bound <= 1:\n" \
                      f"{TAB}{TAB}{TAB}foundWCE = True\n" \
                      f"{TAB}{TAB}{TAB}stats['wce'] = upper_bound\n" \
                      f"{TAB}{TAB}else:\n" \
                      f"{TAB}{TAB}{TAB}lower_bound = stats['et']\n"
        elif self.metric == WHD:
            if_sat += f"{TAB}{TAB}if upper_bound - lower_bound <= 1:\n" \
                      f"{TAB}{TAB}{TAB}foundWCE = True\n" \
                      f"{TAB}{TAB}{TAB}stats['wce'] = upper_bound\n" \
                      f"{TAB}{TAB}else:\n" \
                      f"{TAB}{TAB}{TAB}lower_bound = stats['et']\n"
        elif self.metric == WRE:
            if_sat += f'{TAB}{TAB}if round(upper_bound - lower_bound, 2) <= (10 ** - {self.precision}):\n' \
                      f"{TAB}{TAB}{TAB}foundWCE = True\n" \
                      f"{TAB}{TAB}{TAB}stats['wce'] = upper_bound\n" \
                      f"{TAB}{TAB}else:\n" \
                      f"{TAB}{TAB}{TAB}lower_bound = float(stats['et'])\n"

        if_sat += f"{TAB}{TAB}stats['num_sats'] += 1\n" \
                  f"{TAB}{TAB}stats['sat_runtime'] += (end_iteration - start_iteration)\n" \
                  f"{TAB}{TAB}if stats['et'] == max:\n" \
                  f"{TAB}{TAB}{TAB}foundWCE = True\n" \
                  f"{TAB}{TAB}{TAB}stats['wce'] = stats['et']\n"

        return if_sat

    def express_monotonic_while_loop_sat(self):
        if_sat = ''
        if_sat += f"{TAB}if response == sat:\n" \
                  f"{TAB}{TAB}print(f'sat')\n" \
                  f"{TAB}{TAB}end_iteration = time.time()\n" \
                  f"{TAB}{TAB}returned_model = s.model()\n" \
                  f"{TAB}{TAB}print(f'{{returned_model = }}')\n"
        if self.metric == WAE or self.metric == WRE:
                  f"{TAB}{TAB}print(f\"{{returned_model[f_exact].else_value() = }}\")\n" \
                  f"{TAB}{TAB}print(f\"{{returned_model[f_approx].else_value() = }}\")\n" \
                  f"{TAB}{TAB}print(f\"{{returned_model[f_error].else_value() = }}\")\n"
        elif self.metric == WHD:
            f"{TAB}{TAB}print(f\"{{returned_model[f_error].else_value() = }}\")\n"



        if self.metric == WAE:
            if_sat += f"{TAB}{TAB}returned_value = abs(int(returned_model[f_error].else_value().as_long()))\n" \
                      f"{TAB}{TAB}returned_value_reval = abs(int(returned_model.evaluate(f_error(exact_out, approx_out)).as_long()))\n"
        elif self.metric == WHD:
            if_sat += f"{TAB}{TAB}returned_value = abs(int(returned_model[f_error].else_value().as_long()))\n"

            if_sat +=  f"{TAB}{TAB}returned_value_reval = abs(int(returned_model.evaluate(f_error("
            for i in range(self.graph.num_outputs):
                if i == self.graph.num_outputs - 1:
                    if_sat += f"o{i}_{XOR}_{INT})).as_long()))\n"
                else:
                    if_sat += f"o{i}_{XOR}_{INT}, "




        elif self.metric == WRE:
            if_sat += f"{TAB}{TAB}returned_value = ((returned_model[f_error].else_value().as_decimal({self.precision})))\n" \
                      f"{TAB}{TAB}returned_value_reval = ((returned_model.evaluate(f_error(exact_out, approx_out)).as_decimal({self.precision})))\n"

        if_sat += f"{TAB}{TAB}if returned_value == returned_value_reval:\n" \
                  f"{TAB}{TAB}{TAB}print(f'double-check is passed!')\n" \
                  f"{TAB}{TAB}else:\n" \
                  f"{TAB}{TAB}{TAB}print(f'ERROR!!! double-check failed! exiting...')\n" \
                  f"{TAB}{TAB}{TAB}exit()\n"

        if self.metric == WRE:
            if_sat += f"{TAB}{TAB}if returned_value[-1] == '?':\n" \
                      f"{TAB}{TAB}{TAB}print('removing the last question mark!')\n" \
                      f"{TAB}{TAB}{TAB}returned_value = abs(float(returned_value[:-1])) + 10 ** -({self.precision})\n" \
                      f"{TAB}{TAB}else:\n" \
                      f"{TAB}{TAB}{TAB}returned_value = abs(float(returned_value))\n"

        if self.metric == WAE:
            if_sat += f"{TAB}{TAB}stats['et'] = returned_value\n"
        elif self.metric == WHD:
            if_sat += f"{TAB}{TAB}stats['et'] = returned_value\n"
        elif self.metric == WRE:
            if_sat += f"{TAB}{TAB}stats['et'] = \"{{:.{self.precision}f}}\".format(returned_value)\n"

        if_sat += f"{TAB}{TAB}stats['num_sats'] += 1\n" \
                  f"{TAB}{TAB}stats['sat_runtime'] += (end_iteration - start_iteration)\n" \
                  f"{TAB}{TAB}stats['jumps'].append(returned_value)\n"

        if self.style == 'max':
            if_sat += f"{TAB}{TAB}if stats['et'] == max:\n" \
                        f"{TAB}{TAB}{TAB}foundWCE = True\n" \
                        f"{TAB}{TAB}{TAB}stats['wce'] = stats['et']\n"
        elif self.style == 'min':
            if_sat +=   f"{TAB}{TAB}foundWCE = True\n" \
                        f"{TAB}{TAB}stats['wce'] = stats['et']\n"

        return if_sat

    def express_kind_bisection_while_loop_unsat(self):
        if_unsat = ''

        if_unsat += f"\n" \
                    f"{TAB}if response == unsat:\n" \
                    f"{TAB}{TAB}print('unsat')\n" \
                    f"{TAB}{TAB}end_iteration = time.time()\n"

        if self.metric == WAE:
            if_unsat += f"{TAB}{TAB}if upper_bound - lower_bound <= 1:\n" \
                        f"{TAB}{TAB}{TAB}foundWCE = True\n" \
                        f"{TAB}{TAB}{TAB}if lower_bound == 0:\n" \
                        f"{TAB}{TAB}{TAB}{TAB}stats['wce'] = lower_bound\n" \
                        f"{TAB}{TAB}{TAB}else:\n" \
                        f"{TAB}{TAB}{TAB}{TAB}stats['wce'] = lower_bound\n"
        elif self.metric == WHD:
            if_unsat += f"{TAB}{TAB}if upper_bound - lower_bound <= 1:\n" \
                        f"{TAB}{TAB}{TAB}foundWCE = True\n" \
                        f"{TAB}{TAB}{TAB}if lower_bound == 0:\n" \
                        f"{TAB}{TAB}{TAB}{TAB}stats['wce'] = lower_bound\n" \
                        f"{TAB}{TAB}{TAB}else:\n" \
                        f"{TAB}{TAB}{TAB}{TAB}stats['wce'] = lower_bound\n"
        elif self.metric == WRE:
            if_unsat += f"{TAB}{TAB}if round(upper_bound - lower_bound, 2) <= (10 ** -{self.precision}):\n" \
                        f"{TAB}{TAB}{TAB}foundWCE = True\n" \
                        f"{TAB}{TAB}{TAB}stats['wce'] = lower_bound\n" \
                        f"{TAB}{TAB}{TAB}if lower_bound == 0:\n" \
                        f"{TAB}{TAB}{TAB}{TAB}stats['wce'] = lower_bound\n" \
                        f"{TAB}{TAB}{TAB}else:\n" \
                        f"{TAB}{TAB}{TAB}{TAB}stats['wce'] = lower_bound\n"

        if_unsat += f"{TAB}{TAB}else:\n" \
                    f"{TAB}{TAB}{TAB}upper_bound = float(stats['et'])\n" \
                    f"{TAB}{TAB}stats['num_unsats'] += 1\n" \
                    f"{TAB}{TAB}stats['unsat_runtime'] += (end_iteration - start_iteration)\n" \
                    f"{TAB}s.pop()\n"
        return if_unsat

    def express_bisection_while_loop_unsat(self):
        if_unsat = ''

        if_unsat += f"\n" \
                    f"{TAB}if response == unsat:\n" \
                    f"{TAB}{TAB}print('unsat')\n" \
                    f"{TAB}{TAB}end_iteration = time.time()\n"

        if self.metric == WAE:
            if_unsat += f"{TAB}{TAB}if upper_bound - lower_bound <= 1:\n" \
                        f"{TAB}{TAB}{TAB}foundWCE = True\n" \
                        f"{TAB}{TAB}{TAB}if lower_bound == 0:\n" \
                        f"{TAB}{TAB}{TAB}{TAB}stats['wce'] = lower_bound\n" \
                        f"{TAB}{TAB}{TAB}else:\n" \
                        f"{TAB}{TAB}{TAB}{TAB}stats['wce'] = upper_bound\n"
        elif self.metric == WHD:
            if_unsat += f"{TAB}{TAB}if upper_bound - lower_bound <= 1:\n" \
                        f"{TAB}{TAB}{TAB}foundWCE = True\n" \
                        f"{TAB}{TAB}{TAB}if lower_bound == 0:\n" \
                        f"{TAB}{TAB}{TAB}{TAB}stats['wce'] = lower_bound\n" \
                        f"{TAB}{TAB}{TAB}else:\n" \
                        f"{TAB}{TAB}{TAB}{TAB}stats['wce'] = upper_bound\n"
        elif self.metric == WRE:
            if_unsat += f"{TAB}{TAB}if round(upper_bound - lower_bound, 2) <= (10 ** -{self.precision}):\n" \
                        f"{TAB}{TAB}{TAB}foundWCE = True\n" \
                        f"{TAB}{TAB}{TAB}stats['wce'] = lower_bound\n" \
                        f"{TAB}{TAB}{TAB}if lower_bound == 0:\n" \
                        f"{TAB}{TAB}{TAB}{TAB}stats['wce'] = lower_bound\n" \
                        f"{TAB}{TAB}{TAB}else:\n" \
                        f"{TAB}{TAB}{TAB}{TAB}stats['wce'] = upper_bound\n"

        if_unsat += f"{TAB}{TAB}else:\n" \
                    f"{TAB}{TAB}{TAB}upper_bound = float(stats['et'])\n" \
                    f"{TAB}{TAB}stats['num_unsats'] += 1\n" \
                    f"{TAB}{TAB}stats['unsat_runtime'] += (end_iteration - start_iteration)\n" \
                    f"{TAB}s.pop()\n"
        return if_unsat

    def express_monotonic_while_loop_unsat(self):
        if_unsat = ''
        if_unsat += f"\n" \
                    f"{TAB}if response == unsat:\n" \
                    f"{TAB}{TAB}print('unsat')\n" \
                    f"{TAB}{TAB}end_iteration = time.time()\n" \
                    f"{TAB}{TAB}foundWCE = True\n" \
                    f"{TAB}{TAB}stats['num_unsats'] += 1\n" \
                    f"{TAB}{TAB}stats['unsat_runtime'] += (end_iteration - start_iteration)\n"
        if self.style == 'max':
                    if_unsat +=f"{TAB}{TAB}stats['wce'] = stats['et']\n" \
                               f"{TAB}s.pop()\n"
        elif self.style == 'min':
                    if_unsat += f"{TAB}{TAB}stats['wce'] = 0\n" \
                                f"{TAB}s.pop()\n"

        return if_unsat

    def express_stats(self):
        stats = ''
        stats += f"end_whole = time.time()\n" \
                 f"with open('{self.z3_report}', 'w') as f:\n" \
                 f"{TAB}csvwriter = csv.writer(f)\n" \
                 f"{TAB}header = ['field', 'value']\n" \
                 f"{TAB}csvwriter.writerow(['Experiment', '{self.experiment}'])\n" \
                 f"{TAB}csvwriter.writerow(['WCE', stats['wce']])\n" \
                 f"{TAB}csvwriter.writerow(['Total Runtime', end_whole - start_whole])\n" \
                 f"{TAB}csvwriter.writerow(['SAT Runtime', stats['sat_runtime']])\n" \
                 f"{TAB}csvwriter.writerow(['UNSAT Runtime', stats['unsat_runtime']])\n" \
                 f"{TAB}csvwriter.writerow(['Number of SAT calls', stats['num_sats']])\n" \
                 f"{TAB}csvwriter.writerow(['Number of UNSAT calls', stats['num_unsats']])\n" \
                 f"{TAB}csvwriter.writerow(['Jumps', stats['jumps']])\n"
        return stats

    def express_monotonic_strategy(self):
        monotonic_strategy = ''
        monotonic_strategy += self.declare_stats()
        monotonic_strategy += self.express_monotonic_while_loop()
        return monotonic_strategy

    def express_mc_strategy(self):
        mc_strategy = ''
        mc_strategy += self.declare_stats()
        mc_strategy += self.express_mc_while_loop()
        return mc_strategy

    def express_bisection_strategy(self):
        bisection_strategy = ''
        bisection_strategy += self.declare_stats()
        bisection_strategy += self.express_bisection_while_loop()

        return bisection_strategy

    def express_kind_bisection_strategy(self):
        kind_bisection_strategy = ''
        kind_bisection_strategy += self.declare_stats()
        kind_bisection_strategy += self.express_kind_bisection_while_loop()
        return kind_bisection_strategy

    def export_z3pyscript(self):
        with open(self.out_path, 'w') as z:
            z.writelines(self.z3pyscript)

    def __repr__(self):
        return f'An object of class Z3solver\n' \
               f'{self.name = }\n' \
               f'{self.graph_in_path = }\n' \
               f'{self.out_path = }\n'

    def create_imports(self):
        import_string = f'from z3 import *\n' \
                        f'import sys\n' \
                        f'import time\n' \
                        f'import csv\n' \
                        f'\n'
        return import_string

    def create_abs_function(self):
        abs_function = f'def z3_abs(x):\n' \
                       f'\treturn If(x >= 0, x, -x)\n' \
                       f'\n'

        return abs_function

    def declare_original_circuit(self):
        exact_circuit_declaration = ''
        # inputs
        for n in self.graph.graph.nodes:
            if re.search(r'in\d+', self.graph.graph.nodes[n]['label']):
                exact_circuit_declaration = f'{exact_circuit_declaration}' \
                                            f'{self.declare_gate(n)}\n'
        exact_circuit_declaration += f'\n'
        # gates
        for n in self.graph.graph.nodes:
            if re.search(r'g\d+', n):
                exact_circuit_declaration = f'{exact_circuit_declaration}' \
                                            f'{self.declare_gate(n)}\n'
        # outputs
        for n in self.graph.graph.nodes:
            if re.search(r'out\d+', self.graph.graph.nodes[n]['label']):
                exact_circuit_declaration = f'{exact_circuit_declaration}' \
                                            f'{self.declare_gate(n)}\n'
        return exact_circuit_declaration

    def declare_approximate_circuit(self):
        exact_circuit_declaration = ''
        # inputs
        for n in self.approximate_graph.graph.nodes:
            if self.approximate_graph.is_pruned_pi(n):
                exact_circuit_declaration = f'{exact_circuit_declaration}' \
                                            f'{self.declare_gate(n)}\n'
        exact_circuit_declaration += f'\n'
        # gates
        for n in self.approximate_graph.graph.nodes:
            if re.search(r'g\d+', n):
                exact_circuit_declaration = f'{exact_circuit_declaration}' \
                                            f'{self.declare_gate(n)}\n'
        # outputs
        for n in self.approximate_graph.graph.nodes:
            if re.search(r'out\d+', self.approximate_graph.graph.nodes[n]['label']):
                exact_circuit_declaration = f'{exact_circuit_declaration}' \
                                            f'{self.declare_gate(n)}\n'
        return exact_circuit_declaration

    def declare_gate(self, this_key: str):
        declaration = f"{this_key} = {Z3BOOL}('{this_key}')"
        return declaration

    def express_original_circuit(self):
        exact_circuit_function = ''
        for n in self.graph.graph.nodes:
            if self.graph.is_cleaned_gate(n):
                exact_circuit_function += f'{self.express_exact_gate(n)}\n'
            elif self.graph.is_cleaned_constant(n):
                exact_circuit_function += f'{self.express_exact_constant(n)}\n'
            elif self.graph.is_cleaned_po(n):
                exact_circuit_function += f'{self.express_exact_output(n)}\n'
        exact_circuit_function += f'\n'
        return exact_circuit_function

    def express_approximate_circuit(self):
        approximate_circuit_function = ''

        for n in self.approximate_graph.graph.nodes:
            if self.approximate_graph.is_cleaned_pi(n):
                approximate_circuit_function += f'{self.express_approximate_gate(n)}\n'
            elif self.approximate_graph.is_cleaned_gate(n):
                approximate_circuit_function += f'{self.express_approximate_gate(n)}\n'
            elif self.approximate_graph.is_cleaned_constant(n):
                approximate_circuit_function += f'{self.express_approximate_constant(n)}\n'
            elif self.approximate_graph.is_cleaned_po(n):
                approximate_circuit_function += f'{self.express_approximate_output(n)}\n'
        approximate_circuit_function += f'\n'

        return approximate_circuit_function

    def express_exact_output(self, this_key):
        cur_node = ''
        predecessor = list(self.graph.graph.predecessors(this_key))[0]
        cur_node = f'{this_key} = {predecessor}\n'

        return cur_node

    def express_exact_constant(self, this_key):
        cur_node = ''
        this_constant = re.search(r'TRUE|FALSE', self.graph.graph.nodes[this_key][LABEL]).group()
        cur_node += f'{this_key} = {Z3_GATES_DICTIONARY[this_constant]}\n'
        return cur_node

    def express_exact_gate(self, this_key: str):
        # sth like this: g110 = Not(And(g108, g109))
        cur_node = ''
        if self.graph.is_cleaned_gate(this_key):
            this_gate = re.search(POSSIBLE_GATES, self.graph.graph.nodes[this_key]['label']).group()
            cur_gate = Z3_GATES_DICTIONARY[this_gate]
            predecessor_list = list(self.graph.graph.predecessors(this_key))
            cur_node = f"{this_key}={cur_gate}("
            for idx, u in enumerate(predecessor_list):
                cur_node += f'{u}'
                if idx == len(predecessor_list) - 1:
                    cur_node += f')'
                else:
                    cur_node += f','

        elif self.graph.is_cleaned_pi(this_key):
            pass
        elif self.graph.is_cleaned_po(this_key):
            cur_node = f'{this_key}='
            predecessor_list = list(self.graph.graph.predecessors(this_key))
            for idx, u in enumerate(predecessor_list):
                cur_node += f'{u}'
        return cur_node

    def express_approximate_output(self, this_key):
        cur_node = ''
        predecessor = list(self.approximate_graph.graph.predecessors(this_key))[0]
        cur_node = f'{this_key} = {predecessor}\n'

        return cur_node

    def express_approximate_constant(self, this_key):
        cur_node = ''
        # print(f'{this_key = }')
        this_constant = re.search(r'TRUE|FALSE', self.approximate_graph.graph.nodes[this_key][LABEL]).group()
        cur_node += f'{this_key} = {Z3_GATES_DICTIONARY[this_constant]}\n'
        # if self.approximate_graph.graph.nodes[this_key][PRUNED] == False:
        #     print(f'{this_key = }')
        return cur_node

    def express_approximate_gate(self, this_key: str):
        # sth like this: g110 = Not(And(g108, g109))
        cur_node = ''
        if self.approximate_graph.is_pruned_gate(this_key):
            # cur_node = f'{this_key} = {self.approximate_graph.graph.nodes[this_key][PRUNED]}'
            cur_node = f'# {this_key} is left as a free variable'
            # leave it as free variable
            pass

        elif self.approximate_graph.is_cleaned_gate(this_key):
            this_gate = re.search(POSSIBLE_GATES, self.approximate_graph.graph.nodes[this_key]['label']).group()
            cur_gate = Z3_GATES_DICTIONARY[this_gate]
            predecessor_list = list(self.approximate_graph.graph.predecessors(this_key))
            cur_node = f"{this_key}={cur_gate}("
            for idx, u in enumerate(predecessor_list):
                cur_node += f'{u}'
                if idx == len(predecessor_list) - 1:
                    cur_node += f')'
                else:
                    cur_node += f','

        elif self.approximate_graph.is_pruned_pi(this_key):
            cur_node = f'{this_key}= {self.approximate_graph.graph.nodes[this_key][PRUNED]}'
        elif self.approximate_graph.is_cleaned_po(this_key):
            cur_node = f'{this_key}='
            predecessor_list = list(self.approximate_graph.graph.predecessors(this_key))
            for idx, u in enumerate(predecessor_list):
                cur_node += f'{u}'
        return cur_node

    def declare_original_output(self):
        output_declaration = ''
        # print(f'{self.graph.output_dict = }')



        for i in range(self.graph.num_outputs):
            output_declaration += f"exact_out{i}=Int('exact_out{i}')\n"
            output_declaration += f"exact_out{i}={self.graph.output_dict[i]}*{2 ** i}*2/2\n"

        output_declaration += f"exact_out = Int('exact_out')\n"
        output_declaration += f'exact_out='

        for i in range(self.graph.num_outputs):
            if i == self.graph.num_outputs - 1:
                output_declaration += f'exact_out{i}'
            else:
                output_declaration += f'exact_out{i}+'

        output_declaration += f'\n'
        return output_declaration

    def declare_approximate_output(self):
        output_declaration = ''
        # print(f'{self.approximate_graph.output_dict = }')

        for i in range(self.approximate_graph.num_outputs):
            output_declaration += f"approx_out{i}=Int('approx_out{i}')\n"
            output_declaration += f"approx_out{i}={self.approximate_graph.output_dict[i]}*{2 ** i}*2/2\n"
        # print(f'{self.approximate_graph.output_dict = }')
        output_declaration += f"approx_out = Int('approx_out')\n"
        output_declaration += f'approx_out='
        for i in range(self.approximate_graph.num_outputs):
            if i == self.approximate_graph.num_outputs - 1:
                output_declaration += f'approx_out{i}'
            else:
                output_declaration += f'approx_out{i}+'

        output_declaration += f'\n'
        return output_declaration

    def declare_xor_miter(self):

        xor_miter_declaration = f''
        # o0_xor = Bool('o0_xor')
        # o1_xor = Bool('o1_xor')
        # o2_xor = Bool('o2_xor')
        for i in range(self.graph.num_outputs):
            xor_miter_declaration += f"{MITER}_o{i}_{XOR} = {Z3BOOL}('o{i}_{XOR}')\n"


        # o0_xor = Xor(g27, a27)
        # o1_xor = Xor(g105, a105)
        # o2_xor = Xor(g99, a99)
        for i in range(self.graph.num_outputs):
            xor_miter_declaration += f"o{i}_{XOR} = {Z3XOR}({self.graph.output_dict[i]}, app_{self.graph.output_dict[i]})\n"


        # o0_xor_int = Int('o0_xor_int')
        # o0_xor_int = o0_xor * 2 / 2
        # o1_xor_int = Int('o1_xor_int')
        # o1_xor_int = o1_xor * 2 / 2
        # o2_xor_int = Int('o2_xor_int')
        # o2_xor_int = o2_xor * 2 / 2
        for i in range(self.graph.num_outputs):
            xor_miter_declaration += f"o{i}_{XOR}_{INT} = {Z3INT}('o{i}_{XOR}_{INT}')\n" \
                                     f"o{i}_{XOR}_{INT} = o{i}_{XOR} * 2/2\n"

        return xor_miter_declaration





    def declare_original_function(self):
        exact_function = ''
        exact_function += f'results = []\n'
        exact_function += f"f_exact = Function('f', IntSort(), IntSort())"
        exact_function += f'\n'
        return exact_function

    def declare_approximate_function(self):
        exact_function = ''
        exact_function += f'results = []\n'
        exact_function += f"f_exact = Function('f', IntSort(), IntSort())"
        exact_function += f'\n'
        return exact_function

    def declare_solver(self):
        solver = ''
        if self.optimization == OPTIMIZE or self.optimization == MAXIMIZE:
            solver += f's = Optimize()\n'
        else:
            solver += f's = Solver()\n'

        solver += f's.add(f_exact(exact_out) == exact_out)\n'
        solver += f'\n'

        return solver

    def express_samples(self):
        sample_expression = ''
        for s in self.samples:
            sample_expression += f's.push()\n'
            s_expression = [True if i == '1' else False for i in list(f'{s:0{self.graph.num_inputs}b}')]
            # a 1101 is considered as in3,in2,in1,in0 => in3=1, in2=1, in1=0, in0=1
            s_expression.reverse()  # to keep it consistent with the verilog testbench
            sample_expression += f's.add('
            for idx, e in enumerate(s_expression):
                if idx == len(s_expression) - 1:
                    sample_expression += f'{self.graph.input_dict[idx]}=={e})\n'
                else:
                    sample_expression += f'{self.graph.input_dict[idx]}=={e}, '

            sample_expression += f'sol = s.check()\n'
            sample_expression += f'm = s.model()\n'
            sample_expression += f"print(f'{{m = }}')\n"
            sample_expression += f'cur_result = m[f_exact].else_value().as_string()\n'
            sample_expression += f'results.append(cur_result)\n'

            sample_expression += f's.pop()\n\n'

        return sample_expression

    def store_results(self):
        store_results = ''
        store_results += f"with open('{self.pyscript_results_out_path}', 'w') as f:\n" \
                         f"\tfor line in results:\n" \
                         f"\t\tf.write(line)\n" \
                         f"\t\tf.write('\\n')\n"
        store_results += f"print(f'{{results = }}')"
        return store_results

    # TODO: decorators-----------------------------
    def run_z3pyscript_qor(self):
        with open(self.z3_log_path, 'w') as f:
            process = subprocess.run([PYTHON3, self.out_path],  stderr=PIPE, stdout=PIPE)

    def run_z3pyscript_labeling(self):
        # print(self.pyscript_files_for_labeling )
        for pyscript in self.pyscript_files_for_labeling:
            with open(self.z3_log_path, 'w') as f:
                # process = subprocess.run([PYTHON3, pyscript], stdout=PIPE, stderr=PIPE)
                process = subprocess.run([PYTHON3, pyscript], stderr=PIPE, stdout=PIPE)

                if process.stderr:
                    print(Fore.RED + f'ERROR!!! cannot run {pyscript} properly!')
                    print(f'{process.stderr.decode()}')


    def run_z3pyscript_random(self):
        with open(self.z3_log_path, 'w') as f:
            process = subprocess.run([PYTHON3, self.out_path], stdout=PIPE, stderr=PIPE)

    def run_z3pyscript_test(self):
        with open(self.z3_log_path, 'w') as f:
            # process = subprocess.run([PYTHON3, self.out_path], stdout=PIPE, stderr=PIPE)
            process = subprocess.run([PYTHON3, self.out_path], stderr=PIPE, stdout=PIPE)

        self.set_sample_results(self.import_results())
    # TODO: decorators (end)--------------------------





import re
import subprocess
import os
import shutil
from .config.path import *
from .config.config import *


def get_pure_name(file_name: str) -> str:
    if file_name is None:
        return file_name
    name = file_name
    if re.search('/', file_name):
        name = file_name.split('/')[-1]
    if re.search('\.', name):
        name = name.split('.')[0]
    return name


def fix_direction(file_name: str) -> str:
    file_name = get_pure_name(file_name)
    folder, extension = OUTPUT_PATH['dot']
    dot_in_path = f'{folder}/{file_name}.{extension}'
    folder, extension = OUTPUT_PATH['gv']
    gv_out_path = f'{folder}/{file_name}.{extension}'

    dot_command = f'{DOT} {dot_in_path} -Grankdir=TB -o {gv_out_path}'
    subprocess.call([dot_command], shell=True)


def convert_verilog_to_gv(file_name: str) -> None:
    file_name = get_pure_name(file_name)

    folder, extension = OUTPUT_PATH['ver']
    verilog_in_path = f'{folder}/{file_name}.{extension}'

    folder, extension = OUTPUT_PATH['gv']
    gv_out_path = f'{folder}/{file_name}.{extension}'

    yosys_command = f"""
        read_verilog {verilog_in_path}
        show -prefix {gv_out_path[:-3]} -format gv
        """
    folder, extension = LOG_PATH['yosys']
    log_out_path = f'{folder}/{file_name}_{YOSYS}_v2gv.{extension}'
    with open(f'{log_out_path}', 'w') as y:
        subprocess.call([YOSYS, '-p', yosys_command], stdout=y)
    fix_direction(file_name)


def setup_folder_structure():
    # Setting up the folder structure
    directories = [OUTPUT_PATH['ver'][0], OUTPUT_PATH['aig'][0], OUTPUT_PATH['gv'][0], OUTPUT_PATH['z3'][0],
                   OUTPUT_PATH['report'][0], OUTPUT_PATH['figure'][0], TEST_PATH['tb'][0], TEST_PATH['z3'][0],
                   INPUT_PATH['ver'][0], INPUT_PATH['gv'][0],
                   LOG_PATH['yosys'][0]]

    for directory in directories:
        if ~os.path.exists(directory):
            os.makedirs(directory, exist_ok=True)

def clean_all():
    """
    Cleans the contents of output and test (alongside their subfolders)
    :return: Nothing
    """
    shutil.rmtree('output/')
    shutil.rmtree('test/')




import os
import regex as re
import random
import numpy as np
import subprocess
import sys
from .config.path import *
from .config.config import *
from .argument import Arguments
from .utils import *
from typing import Dict, List


class Verilog:
    def __init__(self, benchmark_name: str, samples: list = []):
        """
        takes in a circuit and creates a verilog object out of it
        :param benchmark_name: the input circuit in verilog format
        :param samples: number of samples for the mc simulation; it is an empty list by default
        """
        self.__circuit_name = get_pure_name(benchmark_name)

        folder, extension = INPUT_PATH['ver']
        self.__circuit_in_path = f'{folder}/{benchmark_name}.{extension}'

        folder, extension = OUTPUT_PATH['ver']
        self.__circuit_out_path = f'{folder}/{benchmark_name}.{extension}'

        folder, extension = OUTPUT_PATH['aig']
        self.__aig_out_path = f'{folder}/{benchmark_name}.{extension}'

        folder, extension = TEST_PATH['tb']
        self.__testbench_out_path = f'{folder}/{benchmark_name}_tb.{extension}'

        folder, extension = TEST_PATH['tb']
        self.__tmp_verilog = f'{folder}/{TMP}.{extension}'

        folder, extension = TEST_PATH['iver']
        self.__iverilog_out_path = f'{folder}/{benchmark_name}.{extension}'

        folder, extension = LOG_PATH['iver']
        self.__iverilog_log_path = f'{folder}/{benchmark_name}_{IVERILOG}_{LOG}.{extension}'

        folder, extension = TEST_PATH['vvp']
        self.__vvp_out_path = f'{folder}/{benchmark_name}.{extension}'

        self.__num_inputs, self.__num_outputs = self.extract_module_io()
        self.__sample_results = None
        self.__samples = samples

        self.synthesize_to_gate_level(self.in_path, self.out_path)
        # self.unwrap_variables()


    @property
    def name(self):
        return self.__circuit_name

    @name.setter
    def name(self, newname):
        self.__circuit_name == newname
    @property
    def num_inputs(self):
        return self.__num_inputs

    @property
    def num_outputs(self):
        return self.__num_outputs

    @property
    def in_path(self):
        return self.__circuit_in_path

    @property
    def out_path(self):
        return self.__circuit_out_path

    @property
    def aig_out_path(self):
        return self.__aig_out_path

    @property
    def cleaned_verilog(self):
        return self.__circuit_out_path

    @property
    def tmp(self):
        return self.__tmp_verilog

    @property
    def testbench(self):
        return self.__testbench_out_path

    @property
    def samples(self):
        return self.__samples

    @property
    def iverilog_out_path(self):
        return self.__iverilog_out_path

    @property
    def iverilog_log_path(self):
        return self.__iverilog_log_path

    @property
    def vvp_in_path(self):
        return self.__iverilog_out_path

    @property
    def vvp_out_path(self):
        return self.__vvp_out_path

    def set_samples(self, samples: np.array or list):
        self.__samples = samples

    @property
    def sample_results(self):
        return self.__sample_results

    def set_sample_results(self, results):
        self.__sample_results = results

    # methods
    def import_results(self):
        arr = []
        with open(self.vvp_out_path, 'r') as f:
            lines = f.readlines()
            for line in lines:
                line = line.replace('\n', '')
                arr.append(int(line, 2))
        results = np.array(arr)

        return results

    def import_circuit(self):
        with open(self.in_path, 'r') as f:
            lines = f.readlines()
        return lines

    def check_multi_vector_declaration(self, cur_list: List[str]) -> List[str]:
        # for lines such as:
        # input [1:0]a, b;
        # cur_list will be = ['[1:0]a', 'b'] which should be ['[1:0]a', '[1:0]b']
        if self.is_vector(cur_list[0]):
            # find the range
            vector_range = re.search('\[\d+:\d+\]', cur_list[0]).group()
            # propagate the range for the rest of the elements of cur_list
            for i in range(1, len(cur_list)):
                cur_list[i] = vector_range + cur_list[i]
        return cur_list

    def is_vector(self, var_name: str) -> bool:
        """
        checks whether var_name variable is a vector
        :param var_name: the variable name
        :return: True if the variable is a vector, otherwise returns False
        """
        if re.search('(\[\d+:\d+\])', var_name):
            return True
        else:
            return False

    def get_name(self, var_name: str) -> str:
        """
        if var_name is an array, e.g., [1:0]a, it will return a. Otherwise, it will return var_name.
        :param var_name: the name of the variable
        :return: a string representing variable name
        """
        if self.is_vector(var_name):
            # remove [n:m] part
            match_obj = re.search('(\[\d+:\d+\])(.*)', var_name)
            return match_obj.group(2)
        else:
            return var_name

    def get_width(self, var_name: str) -> int:
        """
        computes the bit-width of a given variable
        :param var_name: the name of the variable
        :return: an integer representing the bit-width of the given variable
        """
        if self.is_vector(var_name):
            # compute the length
            match = re.search('\[(\d+):(\d+)\]', var_name)  # [1:0]a
            l_bound = int(match.group(1))  # 1
            r_bound = int(match.group(2))  # 0
            return abs((l_bound - r_bound) + 1)
        else:
            return 1

    def extract_inputs_outputs(self, verilog_str: List[str], port_list: List[str]):
        input_dict: Dict[int:(str, int)] = {}
        output_dict: Dict[int:(str, int)] = {}
        # example:
        # for module circuit(a, b, c, d)
        # input [1:0] a;
        # input [2:0]b;
        # output d;
        # output [3:0]c;
        # example input_dict = {0:(a, 2), 1:(b, 3)}
        # example input_dict = {3:(d, 1), 2:(c, 3)}
        for line in verilog_str:
            line = line.strip()  # remove all whitespaces at the beginning or end of the line
            # extract inputs
            if line.startswith('input'):
                match_obj = re.search('input (.+);', line)  # input a, b, c; or input [1:0] a;
                cur_input_str = match_obj.group(1)  # a, b, c or [1:0] a
                cur_input_list = cur_input_str.strip().replace(" ", "").split(',')  # ['a', 'b', 'c'] or ['[1:0]a']
                cur_input_list = self.check_multi_vector_declaration(cur_input_list)
                for inp in cur_input_list:
                    if self.get_name(inp) in port_list:
                        position_in_module_signature = port_list.index(self.get_name(inp))
                        input_dict[position_in_module_signature] = (self.get_name(inp), self.get_width(inp))
                    else:
                        raise Exception(f"input name {self.get_name(inp)} is not in the port_list")
            # extract outputs
            if line.startswith('output'):
                match_obj = re.search('output (.+);', line)  # input a, b, c; or input [1:0] a;
                cur_output_str = match_obj.group(1)  # a, b, c or [1:0] a
                cur_output_list = cur_output_str.strip().replace(" ", "").split(',')  # ['a', 'b', 'c'] or ['[1:0]a']
                cur_output_list = self.check_multi_vector_declaration(cur_output_list)
                for out in cur_output_list:
                    if self.get_name(out) in port_list:
                        position_in_module_signature = port_list.index(self.get_name(out))
                        output_dict[position_in_module_signature] = (self.get_name(out), self.get_width(out))
                    else:
                        raise Exception(f"output name {self.get_name(out)} is not in the port_list")
        return input_dict, output_dict

    def extract_module_signature(self, verilog_str: List[str]):
        """
        reads the first line of a Verilog netlist and extracts module and port names
        :param verilog_str1: the Verilog description as a string
        :return: modulename as a string variable and port names as a list
        """
        module_name = None
        port_list = None
        for line in verilog_str:
            line = line.strip()  # remove whitespaces at the beginning or end of the line
            if line.startswith('module'):
                # extract module
                match_object = re.search('module (\w+)', line)  # module adder(a, b, c)
                module_name = match_object.group(1)  # adder
                # extract port list
                match_object = re.search('.*\((.+)\).*', line)  # module adder(a, b, c)
                ports_str = match_object.group(1)  # a, b, c
                port_list = ports_str.strip().replace(" ", "").split(',')
        assert module_name and port_list, f'Either module_name or port_list is None'
        return module_name, port_list


    def synthesize_to_gate_level(self, input_path: str, output_path: str):
        yosys_command = f"""
        read_verilog {input_path};
        synth -flatten;
        opt;
        opt_clean -purge;
        abc -g NAND;
        opt;
        opt_clean -purge;
        splitnets -ports;
        opt;
        opt_clean -purge;
        write_verilog -noattr {output_path};
        """
        process = subprocess.run(['yosys', '-p', yosys_command], stderr=subprocess.PIPE, stdout=subprocess.PIPE)
        if process.stderr.decode():
            print(f'Error!')
            raise Exception(f'ERROR!!! yosys cannot do its pass on file {input_path}\n{process.stderr.decode()}')
        self.rename_variables(output_path, output_path)



    def rename_variables(self, input_path: str, output_path: str):
        with open(f'{input_path}', 'r') as infile:
            verilog_str = infile.readlines()

        module_name, port_list = self.extract_module_signature(verilog_str)
        input_dict, output_dict = self.extract_inputs_outputs(verilog_str, port_list)

        new_labels = self.create_new_labels(port_list, input_dict, output_dict)
        verilog_str = self.relabel_nodes(verilog_str, new_labels)

        with open(f'{output_path}', 'w') as outfile:
            outfile.writelines(verilog_str)

    def create_new_labels(self, port_list: List, input_dict: Dict, output_dict: Dict):
        new_labels: Dict = {}
        for port_idx in input_dict:
            if input_dict[port_idx][0] == port_list[port_idx]:
                new_labels[port_list[port_idx]] = f'in{port_idx}'
            else:
                raise Exception(f'Error!!! {input_dict[port_idx][0]} is not equal to {port_list[port_idx]}')

        out_idx = 0
        for port_idx in output_dict:
            if output_dict[port_idx][0] == port_list[port_idx]:
                new_labels[port_list[port_idx]] = f'out{out_idx}'
                out_idx += 1
            else:
                raise Exception(f'Error!!! {output_dict[port_idx][0]} is not equal to {port_list[port_idx]}')
        return new_labels

    def relabel_nodes(self, verilog_str: List[str], new_labels: Dict):
        verilog_str_tmp = verilog_str
        for line_idx, line in enumerate(verilog_str):
            for key, value in new_labels.items():
                if key in line:
                    line = line.replace(key, value)
                    verilog_str_tmp[line_idx] = line
        return verilog_str_tmp

    # Deprecated!
    # def synthesize_to_gate_level(self):
    #     yosys_command = f"""
    #     read_verilog {self.in_path}
    #     synth -flatten
    #     opt
    #     opt_clean -purge
    #     abc -g AND,OR
    #     write_verilog -noattr {self.out_path}
    #     abc -g NAND;
    #     aigmap;
    #     opt;
    #     opt_clean -purge;
    #     write_aiger {self.aig_out_path};
    #     """
    #     with open('log.txt', 'w') as y:
    #         subprocess.call([YOSYS, '-p', yosys_command])


    def unwrap_variables(self) -> None:
        """

        :return:
        """
        lsoracle_command = f"""
        read_aig {self.aig_out_path};
        write_verilog {self.out_path}
        """
        with open('lsoracle_log.txt', 'w') as f:
            subprocess.call([LSORACLE, '-c', lsoracle_command])
        self.fix_module_name()

    def fix_module_name(self):
        with open(self.out_path, 'r') as f:
            lines = f.readlines()
        with open(self.out_path, 'w') as g:
            for idx, line in enumerate(lines):
                if re.search('top', line):
                    line = line.replace('top', self.name)
                    lines[idx] = line
                    break
            for line in lines:
                g.write(line)

    def create_test_bench(self, samples: 'list[int]' = []):

        # read a clean verilog
        # create a test bench for it
        num = len(self.samples)
        with open(self.testbench, 'w') as f:
            modulename, port_list, inp, n_inputs, out, n_outputs = self.extract_module_info()

            f.write("module " + modulename + "_tb;\n")
            f.write('reg [' + str(n_inputs - 1) + ':0] pi;\n')
            f.write('wire [' + str(n_outputs - 1) + ':0] po;\n')
            f.write(modulename + ' dut(')

            first = True
            inp_count = 0
            out_count = 0

            for i in port_list:
                if not first:
                    f.write(',')
                first = False

                if i in inp:
                    if inp[i] > 1:
                        f.write(' pi[{}:{}] '.format(inp_count + inp[i] - 1, inp_count))
                    else:
                        f.write(' pi[{}] '.format(inp_count))
                    inp_count += inp[i]

                elif i in out:
                    if out[i] > 1:
                        f.write(' po[{}:{}] '.format(out_count + out[i] - 1, out_count))
                    else:
                        f.write(' po[{}] '.format(out_count))
                    out_count += out[i]

                else:
                    print('[Error] Port {} is not defined as input or output.'.format(i))
                    sys.exit(0)

            f.write(');\n')

            f.write("initial\n")
            f.write("begin\n")

            for idx, sample in enumerate(self.samples):
                
                f.write('# 1  pi=' + str(n_inputs) + '\'b')
                f.write('{0:0>{1}}'.format(str(bin(sample))[2:], n_inputs))
                
                f.write(';\n')
                f.write("#1 $display(\"%b\", po);\n")

            f.write("end\n")
            f.write("endmodule\n")

    def extract_module_io(self) -> (int, int):
        clean_verilog = self.in_path
        yosys_command = 'read_verilog ' + clean_verilog + '; synth -flatten; opt; opt_clean; techmap; write_verilog ' + self.tmp + ';\n'
        process = subprocess.run([YOSYS, '-p', yosys_command], stderr=subprocess.PIPE, stdout=subprocess.PIPE)
        if process.stderr.decode():
            print(f'Error!')
            raise Exception(f'ERROR!!! yosys cannot do its pass on file {clean_verilog}\n{process.stderr.decode()}')
        with open(self.tmp) as tmp_file:
            inp = {}
            inp_count = 0
            out = {}
            out_count = 0
            modulename = None
            line = tmp_file.readline()
            while line:
                tokens = re.split('[ ()]', line.strip().strip(';').strip())

                if len(tokens) > 0 and tokens[0] == 'module' and modulename is None:
                    modulename = tokens[1]
                    port_list = re.split('[,()]', line.strip().strip(';').strip())[1:]
                    port_list = [s.strip() for s in port_list if s.strip() != '']

                if len(tokens) == 2 and (tokens[0] == 'input' or tokens[0] == 'output'):
                    if tokens[0] == 'input':
                        inp[tokens[1]] = 1
                        inp_count += 1
                    if tokens[0] == 'output':
                        out[tokens[1]] = 1
                        out_count += 1

                if len(tokens) == 3 and (tokens[0] == 'input' or tokens[0] == 'output'):
                    range_str = tokens[1][1:-1].split(':')
                    range_int = list(map(int, range_str))
                    length = max(range_int) - min(range_int) + 1
                    if tokens[0] == 'input':
                        inp[tokens[2]] = length
                        inp_count += length
                    if tokens[0] == 'output':
                        out[tokens[2]] = length
                        out_count += length

                line = tmp_file.readline()

        os.remove(self.tmp)

        return inp_count, out_count

    def extract_module_info(self) -> 'tuple(str, list(str), dict, int, dict, int)':
        """
        reads a verilog file and extracts the signature
        :return: a tuple containing the modulename, the list of input and output names, two dictionary that hold the bitwidth
        of all inputs and outputs, and the number of inputs and outputs
        """
        clean_verilog = self.out_path
        yosys_command = 'read_verilog ' + clean_verilog + '; synth -flatten; opt; opt_clean; techmap; write_verilog ' + self.tmp + ';\n'
        subprocess.call([YOSYS, '-p', yosys_command], stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)

        with open(self.tmp) as tmp_file:
            inp = {}
            inp_count = 0
            out = {}
            out_count = 0
            modulename = None
            line = tmp_file.readline()
            while line:
                tokens = re.split('[ ()]', line.strip().strip(';').strip())

                if len(tokens) > 0 and tokens[0] == 'module' and modulename is None:
                    modulename = tokens[1]
                    port_list = re.split('[,()]', line.strip().strip(';').strip())[1:]
                    port_list = [s.strip() for s in port_list if s.strip() != '']

                if len(tokens) == 2 and (tokens[0] == 'input' or tokens[0] == 'output'):
                    if tokens[0] == 'input':
                        inp[tokens[1]] = 1
                        inp_count += 1
                    if tokens[0] == 'output':
                        out[tokens[1]] = 1
                        out_count += 1

                if len(tokens) == 3 and (tokens[0] == 'input' or tokens[0] == 'output'):
                    range_str = tokens[1][1:-1].split(':')
                    range_int = list(map(int, range_str))
                    length = max(range_int) - min(range_int) + 1
                    if tokens[0] == 'input':
                        inp[tokens[2]] = length
                        inp_count += length
                    if tokens[0] == 'output':
                        out[tokens[2]] = length
                        out_count += length

                line = tmp_file.readline()

        os.remove(self.tmp)

        return modulename, port_list, inp, inp_count, out, out_count

    def run_test_bench(self):
        iverilog_command = f'{IVERILOG} -o {self.iverilog_out_path} ' \
                           f'{self.cleaned_verilog} ' \
                           f'{self.testbench} '

        vvp_command = f'{self.iverilog_out_path}'

        with open(f'{self.iverilog_log_path}', 'w') as f:
            subprocess.call(iverilog_command, shell=True, stdout=f)

        with open(f'{self.vvp_out_path}', 'w') as f:
            subprocess.call([VVP, vvp_command], stdout=f)

        os.remove(f'{self.iverilog_out_path}')

        self.set_sample_results(self.import_results())

    def export_circuit(self):
        self.synthesize_to_gate_level(self.in_path, self.out_path)
        # self.unwrap_variables()

    def __repr__(self):
        return f'An object of class Verilog\n' \
               f'{self.name = }\n' \
               f'{self.in_path = }\n' \
               f'{self.out_path = }\n'



import csv
import re
from .config.config import *
from .utils import *

class Stats:
    def __init__(self, file_path: str):
        """
        stores a csv file contents as a an object where each row of the csv file is an attribute
        :param file_path: the path to a csv file
        """
        full_name = get_pure_name(file_path)
        idx = re.search(r'o\d+',full_name)
        s, e = idx.span()
        # print(full_name[:e])
        self.__name = full_name[:e]

        self.__experiment = None
        self.__wce = None
        self.___runtime = None
        self.__sat_runtime = None
        self.__unsat_runtime = None
        self.__num_sats = None
        self.__num_unsats = None
        self.__jumps = None
        self.__file_in_path = file_path
        self.import_csv()

    @property
    def name(self):
        return self.__name

    @property
    def path(self):
        return self.__file_in_path

    @property
    def experiment(self):
        return self.__experiment

    @experiment.setter
    def experiment(self, experiment):
        self.__experiment = experiment

    @property
    def wce(self):
        return self.__wce

    @wce.setter
    def wce(self, wce):
        self.__wce = wce

    @property
    def total_runtime(self):
        return self.___runtime

    @total_runtime.setter
    def total_runtime(self, total_runtime):
        self.___runtime = total_runtime

    @property
    def sat_runtime(self):
        return self.__sat_runtime

    @sat_runtime.setter
    def sat_runtime(self, sat_runtime):
        self.__sat_runtime = sat_runtime

    @property
    def num_sats(self):
        return self.__num_sats

    @num_sats.setter
    def num_sats(self, num_sats):
        self.__num_sats = num_sats

    @property
    def unsat_runtime(self):
        return self.__unsat_runtime

    @unsat_runtime.setter
    def unsat_runtime(self, unsat_runtime):
        self.__unsat_runtime = unsat_runtime

    @property
    def num_unsats(self):
        return self.__num_unsats

    @num_unsats.setter
    def num_unsats(self, num_unsats):
        self.__num_unsats = num_unsats

    @property
    def jump_list(self):
        return self.__jumps

    @jump_list.setter
    def jump_list(self, jump_list):
        self.__jumps = jump_list

    @property
    def file_path(self):
        return self.__file_in_path

    def import_csv(self):
        with open(self.path, 'r') as f:
            csvreader = csv.reader(f)
            for row in csvreader:

                if re.search(EXPERIMENT, row[0]):
                    self.experiment = row[1]
                elif re.search(WCE, row[0]):
                    self.wce = row[1]
                elif re.search(TOTAL_RUNTIME, row[0]):
                    self.total_runtime = row[1]
                elif re.search(SAT_RUNTIME, row[0]):
                    self.sat_runtime = row[1]
                elif re.search(UNSAT_RUNTIME, row[0]):
                    self.unsat_runtime = row[1]
                elif re.search(N_SATS, row[0]):
                    self.num_sats = row[1]
                elif re.search(N_UNSATS, row[0]):
                    self.num_unsats = row[1]
                elif re.search(JUMPS, row[0]):
                    self.jump_list = row[1]


    def __repr__(self):
        return f'An object of class Stats:\n' \
               f'{self.name = }\n' \
               f'{self.experiment = }\n' \
               f'{self.wce = }\n' \
               f'{self.total_runtime = }\n' \
               f'{self.sat_runtime = }\n' \
               f'{self.unsat_runtime}\n' \
               f'{self.num_sats = }\n' \
               f'{self.num_unsats = }\n' \
               f'{self.jump_list = }\n' \
               f''



# tools
YOSYS = 'yosys'
SLANG = 'slang'
LSORACLE = 'lsoracle'
IVERILOG = 'iverilog'
VVP = 'vvp'
DOT = 'dot'

# regex constants
POSSIBLE_GATES=f'and|AND|or|OR|not|NOT'
CONST_1 = f"1'1"
CONST_0 = f"1'0"


# for z3
Z3_GATES_DICTIONARY = {
    'and':'And',
    'AND':'And',
    'or':'Or',
    'OR':'Or',
    'not':'Not',
    'NOT':'Not',
    'T': 'True',
    'True': 'True',
    'true': 'True',
    'TRUE': 'True',
    'F': 'False',
    'False': 'False',
    'false': 'False',
    'FALSE': 'False'
}
REDUNDANT_FIELDS=[
    'height', 'width', 'pos',
    'color', 'fontcolor',
    'rects',
    'contraction', 'contraction',
    'style'
]

# random constants

Z3BOOL = 'Bool'
Z3XOR = 'Xor'
Z3INT = 'Int'

MITER = 'miter'
EXACT = 'exact'
APPROXIMATE = 'approx'
XOR = 'xor'
INT = 'int'

PYTHON3 = 'python3'
Z3 = 'z3py'
SHAPE = 'shape'
LABEL = 'label'
PRUNED = 'pruned'
TAB = '\t'


TMP = 'tmp'
LOG = 'log'

# strategies
MONOTONIC = 'monotonic'
BISECTION = 'bisection'
KIND_BISECTION = 'kind'
MC = 'mc'
DEFAULT_STRATEGY = MONOTONIC

# optimization
OPTIMIZE = 'optimize'
MAXIMIZE = 'maximize'




# Experiments
QOR = 'qor'
RANDOM = 'random'
SINGLE = 'labeling'

# Metrics
WAE = 'wae'
WRE = 'wre'
WHD = 'whd'

# CSV FILE
EXPERIMENT = 'Experiment'
WCE = 'WCE'
TOTAL_RUNTIME = 'Total Runtime'
SAT_RUNTIME = 'SAT Runtime'
UNSAT_RUNTIME = 'UNSAT Runtime'
N_SATS = 'Number of SAT calls'
N_UNSATS = 'Number of UNSAT calls'
JUMPS = 'Jumps'
TEST = 'test'


# PLOTTING
SCATTER = 'scatter_plot'
AVERAGE = 'average_plot'

SMALL_SIZE = 10
MEDIUM_SIZE = 12
BIGGER_SIZE = 15



